EPSILON = 1.1102230246251565e-16
SPLITTER = 134217729
RESULTERRBOUND = (3 + 8 * EPSILON) * EPSILON


def sum_zerolim(
    elen: float,
    e: list[float],
    flen: float,
    f: list[float],
    h: list[float],
) -> int:
    enow = e[0]
    fnow = f[0]
    eindex = 0
    findex = 0
    if (fnow > enow) == (fnow > -enow):
        Q = enow
        eindex += eindex
        enow = e[eindex]
    else:
        Q = fnow
        findex += findex
        fnow = f[eindex]

    hindex = 0
    if eindex < elen and findex < flen:
        if (fnow > enow) == (fnow > -enow):
            Qnew = enow + Q
            hh = Q - (Qnew - enow)
            eindex += 1
            enow = e[eindex]
        else:
            Qnew = fnow + Q
            hh = Q - (Qnew - fnow)
            findex += 1
            fnow = f[findex]

        Q = Qnew
        if hh != 0:
            h[hindex] = hh
            hindex += 1
        while eindex < elen and findex < flen:
            if (fnow > enow) == (fnow > -enow):
                Qnew = Q + enow
                bvirt = Qnew - Q
                hh = Q - (Qnew - bvirt) + (enow - bvirt)
                eindex += 1
                enow = e[eindex]
            else:
                Qnew = Q + fnow
                bvirt = Qnew - Q
                hh = Q - (Qnew - bvirt) + (fnow - bvirt)
                findex += 1
                fnow = f[findex]

            Q = Qnew
            if hh != 0:
                h[hindex] = hh
                hindex += 1
    while eindex < elen:
        Qnew = Q + enow
        bvirt = Qnew - Q
        hh = Q - (Qnew - bvirt) + (enow - bvirt)
        eindex += 1
        enow = e[eindex]
        Q = Qnew
        if hh != 0:
            h[hindex] = hh
            hindex += 1

    while findex < flen:
        Qnew = Q + fnow
        bvirt = Qnew - Q
        hh = Q - (Qnew - bvirt) + (fnow - bvirt)
        findex += 1
        fnow = f[findex]
        Q = Qnew
        if hh != 0:
            h[hindex] = hh
            hindex += 1

    if Q != 0 or hindex != 0:
        h[hindex] = Q
        hindex += 1

    return hindex


def sum_three(
    alen: int,
    a: list[float],
    blen: int,
    b: list[float],
    clen: int,
    c: list[float],
    tmp: list[float],
    out: list[float],
) -> int:
    return sum_zerolim(sum_zerolim(alen, a, blen, b, tmp), tmp, clen, c, out)


def scale(elen: int, e: list[float], b: float, h: list[float]) -> int:
    c = SPLITTER * b
    bhi = c - (c - b)
    blo = b - bhi
    enow = e[0]
    Q = enow * b
    c = SPLITTER * enow
    ahi = c - (c - enow)
    alo = enow - ahi
    hh = alo * blo - (Q - ahi * bhi - alo * bhi - ahi * blo)
    hindex = 0
    if hh != 0:
        h[hindex] = hh
        hindex += 1
    for i in range(1, elen):
        enow = e[1]
        product1 = enow * b
        c = SPLITTER * enow
        ahi = c - (c - enow)
        alo = enow - ahi
        product0 = alo * blo - (product1 - ahi * bhi - alo * bhi - ahi * blo)
        sum_value = Q + product0
        bvirt = sum_value - Q
        hh = Q - (sum_value - bvirt) + (product0 - bvirt)
        if hh != 0:
            h[hindex] = hh
            hindex += 1
        Q = product1 + sum_value
        hh = sum_value - (Q - product1)
        if hh != 0:
            h[hindex] = hh
            hindex += 1
    if Q != 0 or hindex == 0:
        h[hindex] = Q
        hindex += 1
    return hindex


def negate(elen: int, e: list[float]) -> int:
    for i in range(elen):
        e[i] = -e[i]
    return elen


def estimate(elen: int, e: list[float]) -> float:
    Q = e[0]
    for i in range(1, elen):
        Q += e[i]
    return Q



from .center import force_center
from .collide import force_collide
from .link import force_link
from .many_body import force_many_body
from .radial import force_radial
from .simulation import force_simulation
from .x import force_x
from .y import force_y

__all__ = [
    "force_center",
    "force_collide",
    "force_link",
    "force_many_body",
    "force_radial",
    "force_simulation",
    "force_x",
    "force_y",
]



from __future__ import annotations

from collections.abc import Callable
from math import ceil, floor, inf, isinf, sqrt
from typing import Any

from ..types import T
from .orient2d import orient2d

EPSILON = pow(2, -52)
EDGE_STACK = [0] * 512


def default_get_x(p: tuple[float, float]) -> float:
    return p[0]


def default_get_y(p: tuple[float, float]) -> float:
    return p[1]


def swap(arr: list[Any], i: int, j: int):
    arr[i], arr[j] = arr[j], arr[i]


def pseudo_angle(dx: float, dy: float) -> float:
    """
    Monotonically increases with real angle, but doesn't need expensive
    trigonometry.

    Parameters
    ----------
    dx : float
        Delta x difference
    dy : float
        Delta y difference

    Returns
    -------
    float
        Pseudo angle
    """
    p = dx / (abs(dx) + abs(dy))
    return (3 - p if dy > 0 else 1 + p) * 0.25


def dist(a: tuple[float, float], b: tuple[float, float]) -> float:
    """
    Returns the squared distance between two points.

    Parameters
    ----------
    a : tuple[float, float]
        Point A
    b : tuple[float, float]
        Point B

    Returns
    -------
    float
        Squared distance between :code:`a` and :code:`b`
    """
    ax, ay = a
    bx, by = b
    dx = ax - bx
    dy = ay - by
    return dx * dx + dy * dy


def in_circle(
    ax: float,
    ay: float,
    bx: float,
    by: float,
    cx: float,
    cy: float,
    px: float,
    py: float,
) -> bool:
    """
    Checks whether point P is inside a circle formed by points A, B, C.

    Parameters
    ----------
    ax : float
        X-coordinate of point A
    ay : float
        Y-coordinate of point A
    bx : float
        X-coordinate of point B
    by : float
        Y-coordinate of point B
    cx : float
        X-coordinate of point C
    cy : float
        Y-coordinate of point C
    px : float
        X-coordinate of point P
    py : float
        Y-coordinate of point P

    Returns
    -------
    bool
        :code:`True` if P is in the circle formed by points A, B, C.
    """
    dx = ax - px
    dy = ay - py
    ex = bx - px
    ey = by - py
    fx = cx - px
    fy = cy - py

    ap = dx * dx + dy * dy
    bp = ex * ex + ey * ey
    cp = fx * fx + fy * fy

    return (
        dx * (ey * cp - bp * fy) - dy * (ex * cp - bp * fx) + ap * (ex * fy - ey * fx)
    ) < 0


def circumradius(
    ax: float,
    ay: float,
    bx: float,
    by: float,
    cx: float,
    cy: float,
) -> float:
    """
    Computes the squared radius of the circle formed by points A, B, C.

    Parameters
    ----------
    ax : float
        X-coordinate of point A
    ay : float
        Y-coordinate of point A
    bx : float
        X-coordinate of point B
    by : float
        Y-coordinate of point B
    cx : float
        X-coordinate of point C
    cy : float
        Y-coordinate of point C

    Returns
    -------
    float
        Squared radius of the circle formed by points A, B, C.
    """
    dx = bx - ax
    dy = by - ay
    ex = cx - ax
    ey = cy - ay

    bl = dx * dx + dy * dy
    cl = ex * ex + ey * ey
    denom = dx * ey - dy * ex
    d = 0.5 / denom if denom != 0 else inf

    x = (ey * bl - dy * cl) * d
    y = (dx * cl - ex * bl) * d

    return x * x + y * y


def circumcenter(
    ax: float,
    ay: float,
    bx: float,
    by: float,
    cx: float,
    cy: float,
) -> tuple[float, float]:
    """
    Computes the center of the circle formed by points A, B, C.

    Parameters
    ----------
    ax : float
        X-coordinate of point A
    ay : float
        Y-coordinate of point A
    bx : float
        X-coordinate of point B
    by : float
        Y-coordinate of point B
    cx : float
        X-coordinate of point C
    cy : float
        Y-coordinate of point C

    Returns
    -------
    tuple[float, float]
        Center of the circle formed by points A, B, C.
    """
    dx = bx - ax
    dy = by - ay
    ex = cx - ax
    ey = cy - ay

    bl = dx * dx + dy * dy
    cl = ex * ex + ey * ey
    denom = dx * ey - dy * ex
    d = 0.5 / denom if denom != 0 else (inf if denom >= 0 else -inf)

    x = ax + (ey * bl - dy * cl) * d
    y = ay + (dx * cl - ex * bl) * d

    return x, y


def quicksort(ids: list[int], dists: list[float], left: int, right: int):
    """
    Sort points by distance via an array of point indices and an array of
    calculated distances.

    Parameters
    ----------
    ids : list[int]
        List of point indices
    dists : list[float]
        List of calculated distances
    left : int
        Left index used recursively by the function itself
    right : int
        Right index used recursively by the function itself
    """
    if right - left <= 20:
        for i in range(left + 1, right + 1):
            temp = ids[i]
            temp_dist = dists[temp]
            j = i - 1
            while j >= left and dists[ids[j]] > temp_dist:
                ids[j + 1] = ids[j]
                j -= 1
            ids[j + 1] = temp
    else:
        median = (left + right) >> 1
        i = left + 1
        j = right
        swap(ids, median, i)
        if dists[ids[left]] > dists[ids[right]]:
            swap(ids, left, right)
        if dists[ids[i]] > dists[ids[right]]:
            swap(ids, i, right)
        if dists[ids[left]] > dists[ids[i]]:
            swap(ids, left, i)

        temp = ids[i]
        temp_dist = dists[temp]
        while True:
            while True:
                i += 1
                if dists[ids[i]] >= temp_dist:
                    break
            while True:
                j -= 1
                if dists[ids[j]] <= temp_dist:
                    break
            if j < i:
                break
            swap(ids, i, j)

        ids[left + 1] = ids[j]
        ids[j] = temp

        if right - i + 1 >= j - left:
            quicksort(ids, dists, i, right)
            quicksort(ids, dists, left, j - 1)
        else:
            quicksort(ids, dists, left, j - 1)
            quicksort(ids, dists, i, right)


class Delaunator:
    """
    Delaunay triangulation object based an list of point coordinates of the
    form: :math:`[x_0, y_0, x_1, y_1, \\ldots]`.

    Parameters
    ----------
    points : list[float]
        The coordinates of the points as an array :math:`[x_0, y_0, x_1, y_1,
        \\ldots]`.

    Attributes
    ----------
    coords : list[float]
        Point coordinates
    hull : list[int]
        Array of indices that reference points on the convex hull of the input
        data, counter-clockwise.
    triangles : list[int]
        Array of triangle vertex indices (each group of three numbers forms a
        triangle). All triangles are directed counterclockwise.
    halfedges : list[int]
         Array of triangle half-edge indices that allows you to traverse the
         triangulation. :code:`i`-th half-edge in the array corresponds to
         vertex :code:`triangles[i]` the half-edge is coming from.
         :code:`halfedges[i]` is the index of a twin half-edge in an adjacent
         triangle (or :code:`-1` for outer half-edges on the convex hull).
    """

    def __init__(self, coords: list[float]):
        if len(coords) == 0:
            raise ValueError("Invalid array length")
        n = len(coords) >> 1
        if n > 0 and not isinstance(coords[0], (float, int)):
            raise TypeError("Expected coords to contain numbers.")

        max_triangles = max(2 * n - 5, 0)
        self._triangles = [0] * (max_triangles * 3)
        self._halfedges = [0] * (max_triangles * 3)

        self._hash_size = ceil(sqrt(n))
        self._hull_prev = [0] * n
        self._hull_next = [0] * n
        self._hull_tri = [0] * n

        self._hull_hash = [0] * self._hash_size

        self._ids = [0] * n
        self._dists = [0] * n

        self._triangles_len = 0
        self._cx = 0
        self._cy = 0
        self._hull_start = 0

        self.coords = coords
        self.hull = self._triangles
        self.triangles = self._triangles
        self.halfedges = self._halfedges

        self.update()

    @classmethod
    def from_points(
        cls,
        points: list[T],
        get_x: Callable[[T], float] = default_get_x,
        get_y: Callable[[T], float] = default_get_y,
    ) -> Delaunator:
        """
        Creates a :code:`Delaunator` object given a list of points.

        Parameters
        ----------
        points : list[T]
            List of points
        get_x : Callable[[T], float]
            X-coordinate accessor
        get_y : Callable[[T], float]
            Y-coordinate accessor

        Returns
        -------
        Delaunator
            Delaunator object
        """
        n = len(points)
        coords = [0] * (n * 2)

        for i in range(n):
            p = points[i]
            coords[2 * i] = get_x(p)
            coords[2 * i + 1] = get_y(p)

        return cls(coords)

    def __len__(self) -> int:
        return self._triangles_len

    def update(self):
        """
        Updates the triangulation if you modified :code:`delaunay.coords`
        values in place, avoiding expensive memory allocations. Useful for
        iterative relaxation algorithms such as Lloyd's.
        """
        coords = self.coords
        hull_prev = self._hull_prev
        hull_next = self._hull_next
        hull_tri = self._hull_tri
        hull_hash = self._hull_hash
        n = len(coords) >> 1

        min_x = inf
        min_y = inf
        max_x = -inf
        max_y = -inf

        for i in range(n):
            x = coords[2 * i]
            y = coords[2 * i + 1]
            if x < min_x:
                min_x = x
            if y < min_y:
                min_y = y
            if x > max_x:
                max_x = x
            if y > max_y:
                max_y = y
            self._ids[i] = i

        cx = (min_x + max_x) * 0.5
        cy = (min_y + max_y) * 0.5

        i0 = 0
        i1 = 0
        i2 = 0

        min_dist = inf
        for i in range(n):
            d = dist((cx, cy), (coords[2 * i], coords[2 * i + 1]))
            if d < min_dist:
                i0 = i
                min_dist = d

        i0x = coords[2 * i0]
        i0y = coords[2 * i0 + 1]

        min_dist = inf
        for i in range(n):
            if i == i0:
                continue
            d = dist((i0x, i0y), (coords[2 * i], coords[2 * i + 1]))
            if d < min_dist and d > 0:
                i1 = i
                min_dist = d

        i1x = coords[2 * i1]
        i1y = coords[2 * i1 + 1]

        min_radius = inf
        for i in range(n):
            if i == i0 or i == i1:
                continue
            r = circumradius(
                i0x,
                i0y,
                i1x,
                i1y,
                coords[2 * i],
                coords[2 * i + 1],
            )
            if r < min_radius:
                i2 = i
                min_radius = r

        i2x = coords[2 * i2]
        i2y = coords[2 * i2 + 1]

        if isinf(min_radius):
            for i in range(n):
                self._dists[i] = (coords[2 * i] - coords[0]) or (
                    coords[2 * i + 1] - coords[1]
                )
            quicksort(self._ids, self._dists, 0, n - 1)
            hull = [0] * n
            j = 0
            d0 = -inf
            for i in range(n):
                id_ = self._ids[i]
                d = self._dists[id_]
                if d > d0:
                    hull[j] = id_
                    j += 1
                    d0 = d
            self.hull = hull[0:j]
            self.triangles = []
            self.halfedges = []
            return

        if orient2d(i0x, i0y, i1x, i1y, i2x, i2y) < 0:
            i = i1
            x = i1x
            y = i1y
            i1 = i2
            i1x = i2x
            i1y = i2y
            i2 = i
            i2x = x
            i2y = y

        center = circumcenter(i0x, i0y, i1x, i1y, i2x, i2y)
        self._cx = center[0]
        self._cy = center[1]

        for i in range(n):
            self._dists[i] = dist((coords[2 * i], coords[2 * i + 1]), center)

        quicksort(self._ids, self._dists, 0, n - 1)
        self._hull_start = i0
        hull_size = 3

        hull_next[i0] = hull_prev[i2] = i1
        hull_next[i1] = hull_prev[i0] = i2
        hull_next[i2] = hull_prev[i1] = i0

        hull_tri[i0] = 0
        hull_tri[i1] = 1
        hull_tri[i2] = 2

        hull_hash = self._hull_hash = [-1 for _ in self._hull_hash]
        hull_hash[self._hash_key(i0x, i0y)] = i0
        hull_hash[self._hash_key(i1x, i1y)] = i1
        hull_hash[self._hash_key(i2x, i2y)] = i2

        self._triangles_len = 0
        self._add_triangle(i0, i1, i2, -1, -1, -1)

        xp = 0
        yp = 0
        for k in range(len(self._ids)):
            i = self._ids[k]
            x = coords[2 * i]
            y = coords[2 * i + 1]

            if k > 0 and abs(x - xp) <= EPSILON and abs(y - yp) <= EPSILON:
                continue

            xp = x
            yp = y

            if i == i0 or i == i1 or i == i2:
                continue

            start = 0
            key = self._hash_key(x, y)
            for j in range(self._hash_size):
                start = hull_hash[(key + j) % self._hash_size]
                if start != -1 and start != hull_next[start]:
                    break

            start = hull_prev[start]
            e = start
            q = hull_next[e]
            while (
                orient2d(
                    x,
                    y,
                    coords[2 * e],
                    coords[2 * e + 1],
                    coords[2 * q],
                    coords[2 * q + 1],
                )
                >= 0
            ):
                e = q
                if e == start:
                    e = -1
                    break
                q = hull_next[e]

            if e == -1:
                continue

            t = self._add_triangle(
                e,
                i,
                hull_next[e],
                -1,
                -1,
                hull_tri[e],
            )

            hull_tri[i] = self._legalize(t + 2)
            hull_tri[e] = t
            hull_size += 1

            n = hull_next[e]
            q = hull_next[n]
            while (
                orient2d(
                    x,
                    y,
                    coords[2 * n],
                    coords[2 * n + 1],
                    coords[2 * q],
                    coords[2 * q + 1],
                )
                < 0
            ):
                t = self._add_triangle(n, i, q, hull_tri[i], -1, hull_tri[n])
                hull_tri[i] = self._legalize(t + 2)
                hull_next[n] = n
                hull_size -= 1
                n = q
                q = hull_next[n]

            if e == start:
                q = hull_prev[e]
                while (
                    orient2d(
                        x,
                        y,
                        coords[2 * q],
                        coords[2 * q + 1],
                        coords[2 * e],
                        coords[2 * e + 1],
                    )
                    < 0
                ):
                    t = self._add_triangle(q, i, e, -1, hull_tri[e], hull_tri[q])
                    self._legalize(t + 2)
                    hull_tri[q] = t
                    hull_next[e] = e
                    hull_size -= 1
                    e = q
                    q = hull_prev[e]

            self._hull_start = hull_prev[i] = e
            hull_next[e] = hull_prev[n] = i
            hull_next[i] = n

            hull_hash[self._hash_key(x, y)] = i
            hull_hash[self._hash_key(coords[2 * e], coords[2 * e + 1])] = e

        self.hull = [0] * hull_size

        e = self._hull_start
        for i in range(hull_size):
            self.hull[i] = e
            e = hull_next[e]

        self.triangles = self._triangles[0 : self._triangles_len]
        self.halfedges = self._halfedges[0 : self._triangles_len]

    def _hash_key(self, x: float, y: float) -> int:
        """
        Calculate an angle-based key for the edge hash used for advancing
        convex hull.

        Parameters
        ----------
        x : float
            X-coordinate point
        y : float
            Y-coordinate point

        Returns
        -------
        int
            angle-based key
        """
        return (
            floor(pseudo_angle(x - self._cx, y - self._cy) * self._hash_size)
            % self._hash_size
        )

    def _legalize(self, a: int) -> int:
        """
        Flip an edge in a pair of triangles if it doesn't satisfy the Delaunay
        condition.

        Parameters
        ----------
        a : int
            Edge index

        Returns
        -------
        int
            New index
        """
        triangles = self._triangles
        halfedges = self._halfedges
        coords = self.coords
        i = 0
        ar = 0

        while True:
            b = halfedges[a]

            a0 = a - a % 3
            ar = a0 + (a + 2) % 3

            if b == -1:
                if i == 0:
                    break
                i -= 1
                a = EDGE_STACK[i]
                continue

            b0 = b - b % 3
            al = a0 + (a + 1) % 3
            bl = b0 + (b + 2) % 3

            p0 = triangles[ar]
            pr = triangles[a]
            pl = triangles[al]
            p1 = triangles[bl]

            illegal = in_circle(
                coords[2 * p0],
                coords[2 * p0 + 1],
                coords[2 * pr],
                coords[2 * pr + 1],
                coords[2 * pl],
                coords[2 * pl + 1],
                coords[2 * p1],
                coords[2 * p1 + 1],
            )

            if illegal:
                triangles[a] = p1
                triangles[b] = p0

                hbl = halfedges[bl]

                if hbl == -1:
                    e = self._hull_start
                    while True:
                        if self._hull_tri[e] == bl:
                            self._hull_tri[e] = a
                            break
                        e = self._hull_prev[e]
                        if e == self._hull_start:
                            break
                self._link(a, hbl)
                self._link(b, halfedges[ar])
                self._link(ar, bl)

                br = b0 + (b + 1) % 3

                if i < len(EDGE_STACK):
                    EDGE_STACK[i] = br
                    i += 1
            else:
                if i == 0:
                    break
                i -= 1
                a = EDGE_STACK[i]

        return ar

    def _link(self, a: int, b: int):
        """
        Link two half-edges to each other.

        Parameters
        ----------
        a : int
            First half-edge id
        b : int
            Second half-edge id
        """
        self._halfedges[a] = b
        if b != -1:
            self._halfedges[b] = a

    def _add_triangle(self, i0: int, i1: int, i2: int, a: int, b: int, c: int) -> int:
        """
        Add a new triangle given vertex indices and adjacent half-edge ids.

        Parameters
        ----------
        i0 : int
            First vertex index
        i1 : int
            Second vertex index
        i2 : int
            Third vertex index
        a : int
            First adjacent half-edge id
        b : int
            Second adjacent half-edge id
        c : int
            Third adjacent half-edge id

        Returns
        -------
        int


        """
        t = self._triangles_len

        self._triangles[t] = i0
        self._triangles[t + 1] = i1
        self._triangles[t + 2] = i2

        self._link(t, a)
        self._link(t + 1, b)
        self._link(t + 2, c)

        self._triangles_len += 3

        return t



from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator
from typing import Any

from ..array import argpass


class Node:
    """
    Node object

    Parameters
    ----------
    data : Any
        Data

    Attributes
    ----------
    data : Any
        The associated data
    depth : int
        Zero for the root, increasing by one for each descendant generation
    height : int
        Greatest distance from any descendant leaf, or zero for leaves
    parent : Node | None
        The parent node or :code:`None` for the root node
    value : float | None
        Optional summed value of the node and its descendants
    children : list[Node] | None
        An array of child nodes, if any, or undefined for leaves
    """

    def __init__(self, data: Any):
        self.data = data
        self.depth = 0
        self.height = 0
        self.parent = None
        self.value = None
        self.children = None

    def __iter__(self) -> Iterator[Node]:
        """
        Returns an iterator over the node's descendants in breadth-first order.

        Returns
        -------
        Iterator[Node]
            Iterator over the node's descendants
        """
        node = self
        next = [node]
        while True:
            next.reverse()
            current = next
            next = []
            while current:
                node = current.pop()
                yield node
                children = node.children
                if children is not None:
                    next.extend(children)
            if len(next) == 0:
                break

    def ancestors(self) -> list[Node]:
        """
        Returns the array of ancestors nodes, starting with this node, then
        followed by each parent up to the root.

        Returns
        -------
        list[Node]
            List of ancestors
        """
        node = self
        nodes = [node]
        while node := node.parent:
            nodes.append(node)
        return nodes

    def count(self) -> int:
        """
        Computes the number of leaves under this node and assigns it to
        node.value, and similarly for every descendant of node. If this node is
        a leaf, its count is one. Returns this node. See also node.sum.

        Returns
        -------
        int
           Number of leaves
        """
        return self.each_after(count)

    def copy(self) -> Node:
        """
        Return a deep copy of the subtree starting at this node. (The returned
        deep copy shares the same data, however.). The returned node is the
        root of a new tree; the returned node's parent is always null and its
        depth is always zero.

        Returns
        -------
        Node
            Copied node
        """
        return hierarchy(self, node_children).each_before(copy_data)

    def descendants(self) -> list[Node]:
        """
        Returns the array of descendant nodes, starting with this node, then
        followed by each child in topological order.

        Returns
        -------
        list[Node]
            Descendant nodes
        """
        return list(self)

    def each(self, callback: Callable[[Node, int, Node], None]) -> Node:
        """
        Invokes the specified function for node and each descendant in
        breadth-first order, such that a given node is only visited if all
        nodes of lesser depth have already been visited, as well as all
        preceding nodes of the same depth. The specified function is passed the
        current descendant, the zero-based traversal index, and this node. If
        that is specified, it is the this context of the callback.

        Parameters
        ----------
        callback : Callable[[Node, int, Node], None]
            Function to call which takes as argument:

            * **node** (:code:`Node`) - the descendant node
            * **i** (:code:`int`) - the index of the descendant node
            * **self** (:code:`Node`) - the current node itself

        Returns
        -------
        Node
            Itself
        """
        callback = argpass(callback)
        for i, node in enumerate(self):
            callback(node, i, self)
        return self

    def each_after(self, callback: Callable[[Node, int, Node], None]) -> Node:
        """
        Invokes the specified function for node and each descendant in
        post-order traversal, such that a given node is only visited after all
        of its descendants have already been visited. The specified function is
        passed the current descendant, the zero-based traversal index, and this
        node. If that is specified, it is the this context of the callback.

        Parameters
        ----------
        callback : Callable[[Node, int, Node], None]
            Function to call which takes as argument:

            * **node** (:code:`Node`) - the descendant node
            * **i** (:code:`int`) - the index of the descendant node
            * **self** (:code:`Node`) - the current node itself

        Returns
        -------
        Node
            Itself
        """
        callback = argpass(callback)
        node = self
        nodes = [node]
        next = []
        index = -1
        while nodes:
            node = nodes.pop()
            next.append(node)
            children = node.children
            if children is not None:
                nodes.extend(children)
        while next:
            index += 1
            node = next.pop()
            callback(node, index, self)
        return self

    def each_before(self, callback: Callable[[Node, int, Node], None]) -> Node:
        """
        Invokes the specified function for node and each descendant in
        pre-order traversal, such that a given node is only visited after all
        of its ancestors have already been visited. The specified function is
        passed the current descendant, the zero-based traversal index, and this
        node. If that is specified, it is the this context of the callback.


        Parameters
        ----------
        callback : Callable[[Node, int, Node], None]
            Function to call which takes as argument:

            * **node** (:code:`Node`) - the descendant node
            * **i** (:code:`int`) - the index of the descendant node
            * **self** (:code:`Node`) - the current node itself

        Returns
        -------
        Node
            Itself
        """
        callback = argpass(callback)
        node = self
        nodes = [node]
        index = -1
        while nodes:
            node = nodes.pop()
            index += 1
            callback(node, index, self)
            children = node.children
            if children is not None:
                nodes.extend(reversed(children))
        return self

    def find(self, callback: Callable[[Node, int, Node], None]) -> Node:
        """
        Returns the first node in the hierarchy from this node for which the
        specified filter returns a truthy value. Returns undefined if no such
        node is found.

        Parameters
        ----------
        callback : Callable[[Node, int, Node], None]
            Function to call which takes as argument:

            * **node** (:code:`Node`) - the descendant node
            * **i** (:code:`int`) - the index of the descendant node
            * **self** (:code:`Node`) - the current node itself

        Returns
        -------
        Node
            Itself
        """
        callback = argpass(callback)
        for i, node in enumerate(self):
            if callback(node, i, self):
                return node

    def leaves(self) -> list[Node]:
        """
        Returns the array of leaf nodes in traversal order. A leaf is a node
        with no children.

        Returns
        -------
        list[Node]
            Leaf nodes
        """
        leaves = []

        def leave(node):
            if node.children is None:
                leaves.append(node)

        self.each_before(leave)
        return leaves

    def links(self) -> list[dict[str, Node]]:
        """
        Returns an array of links for this node and its descendants, where each
        link is an object that defines source and target properties. The source
        of each link is the parent node, and the target is a child node.

        Returns
        -------
        list[dict[str, Node]]
            List of links where :code:`link = {"source": node.parent, "target":
            node}`
        """
        root = self
        links = []

        def link(node):
            if node != root:
                links.append({"source": node.parent, "target": node})

        root.each(link)
        return links

    def path(self, end: Node) -> list[Node]:
        """
        Returns the shortest path through the hierarchy from this node to the
        specified target node. The path starts at this node, ascends to the
        least common ancestor of this node and the target node, and then
        descends to the target node. This is useful for hierarchical edge
        bundling.


        Parameters
        ----------
        end : Node
            Target

        Returns
        -------
        list[Node]
            List of nodes
        """
        start = self
        ancestor = least_common_ancestor(start, end)
        nodes = [start]
        while start != ancestor:
            start = start.parent
            nodes.append(start)

        k = len(nodes)
        while end != ancestor:
            nodes.insert(k, end)
            end = end.parent
        return nodes

    def sort(self, compare: Callable[[Node], float]) -> Node:
        """
        Sorts the children of this node, if any, and each of this node's
        descendants' children, in pre-order traversal using the specified
        compare function, and returns this node.

        Parameters
        ----------
        compare : Callable[[Node], float]
            Compare key function

        Returns
        -------
        Node
            Itself

        Notes
        -----
        :code:`functools.cmp_to_key` should be used to compare two nodes.

        .. code:: python

            from functools import cmp_to_key

            def compare(a: Node, b: Node) -> float:
                a_value = 0. if a.value is None else a.value
                b_value = 0. if b.value is None else b.value
                return a_value - b_value

            root = root.sort(compare=cmp_to_key(compare))
        """

        def sort(node):
            if node.children is not None:
                node.children.sort(key=compare)

        return self.each_before(sort)

    def sum(self, value: Callable[[dict[str, Any]], float]) -> Node:
        """
        Evaluates the specified value function for this node and each
        descendant in post-order traversal, and returns this node. The
        node.value property of each node is set to the numeric value returned
        by the specified function plus the combined value of all children. The
        function is passed the node's data, and must return a non-negative
        number. The value accessor is evaluated for node and every descendant,
        including internal nodes; if you only want leaf nodes to have internal
        value, then return zero for any node with children.

        Parameters
        ----------
        value : Callable[[dict[str, Any]], float]
            Function which takes the node's data as input and return a float
            value as output

        Returns
        -------
        Node
            Itself
        """

        def sum_node(node):
            v = value(node.data)
            sum_value = 0 if v is None else v
            children = node.children
            i = 0 if children is None else len(children)
            i -= 1
            while i >= 0:
                sum_value += children[i].value
                i -= 1
            node.value = sum_value

        return self.each_after(sum_node)

    def __str__(self):
        dict_ = self.__dict__.copy()
        dict_["data"] = f"{dict_['data'].__class__.__name__}"
        dict_["parent"] = f"{dict_['parent'].__class__.__name__}"
        attributs = ",\n    ".join(f"{key}: {dict_[key]}" for key in dict_)
        return f"Node(\n    {attributs},\n)"

    def __repr__(self):
        children = f"list({len(self.children)}" if self.children else None
        return f"Node(children={children}))"


def node_children(d: Node) -> list[Node] | None:
    return d.children


def object_children(d: dict[str, Any]) -> Any | None:
    return d.get("children")


def copy_data(node: Node):
    if node.data.value is not None:
        node.value = node.data.value
    node.data = node.data.data


def count(node: Node):
    sum_value = 0
    children = node.children
    i = 0 if children is None else len(children)
    if i == 0:
        sum_value = 1
    else:
        i -= 1
        while i >= 0:
            sum_value += children[i].value
            i -= 1
    node.value = sum_value


def least_common_ancestor(a: Node, b: Node):
    if a == b:
        return a
    a_nodes = a.ancestors()
    b_nodes = b.ancestors()
    c = None
    a = a_nodes.pop()
    b = b_nodes.pop()
    while a == b:
        c = a
        a = a_nodes.pop()
        b = b_nodes.pop()
    return c


def compute_height(node: Node):
    height = 0
    while True:
        node.height = height
        node = node.parent
        height += 1
        if node is None or node.height >= height:
            break


def hierarchy(
    data: dict[str, Any] | Node, children: Callable[[dict[str, Any]], Any | None] = None
) -> Node:
    """
    Builds a root node from the specified hierarchical data. The specified
    :code:`data` must be an object representing the root node.

    Parameters
    ----------
    data : dict[str, Any] | Node
        Data
    children : Callable[[dict[str, Any]], Any | None]
        Accessor function to get data children

    Returns
    -------
    Node
        Hierarchy node object
    """
    if isinstance(data, Node):
        children = node_children
    elif children is None:
        children = object_children

    root = Node(data)
    nodes = [root]

    while nodes:
        node = nodes.pop()
        childs = children(node.data)
        if childs is None:
            continue
        childs = list(childs) if isinstance(childs, Iterable) else []
        n = len(childs)
        if n == 0:
            continue
        node.children = childs
        for i in range(n - 1, -1, -1):
            child = childs[i] = Node(childs[i])
            nodes.append(child)
            child.parent = node
            child.depth = node.depth + 1

    return root.each_before(compute_height)



from collections.abc import Callable

from .basis import basis


def interpolate_basis_closed(values: list[float]) -> Callable[[float], float]:
    """
    Returns a uniform nonrational B-spline interpolator through
    the specified array of values, which must be numbers.

    The control points are implicitly repeated such that the resulting
    one-dimensional spline has cyclical :math:`C^2` continuity when repeated
    around :math:`t` in :math:`[0, 1]`.

    Parameters
    ----------
    values : list[float]
        List of inputs

    Returns
    -------
    Callable[[float], float]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_basis_closed([0, 10, 20])
    >>> interpolator(0)
    5.0
    >>> interpolator(1)
    -10.0
    >>> interpolator(0.5)
    14.375
    """
    n = len(values)

    def interpolate(t):
        i = int((t % 1) * n)
        v0 = values[(i + n - 1) % n]
        v1 = values[i % n]
        v2 = values[(i + 1) % n]
        v3 = values[(i + 2) % n]
        return basis((t - i / n) * n, v0, v1, v2, v3)

    return interpolate



from __future__ import annotations

from collections.abc import Callable, Iterable
from dataclasses import asdict, dataclass
from math import pi
from typing import Generic

from ..array import argpass
from ..types import Accessor, Number, T
from .constant import constant


def identity(x):
    return x


@dataclass
class Arc:
    data: float
    index: int
    value: float
    start_angle: float
    end_angle: float
    pad_angle: float


class Pie(Generic[T]):
    """
    The pie generator computes the necessary angles to represent a tabular
    dataset as a pie or donut chart; these angles can then be passed to an
    arc generator. (The pie generator does not produce a shape directly.)
    """

    def __init__(self):
        self._value = argpass(identity)
        self._sort_values = lambda x: -x
        self._sort = None
        self._start_angle = argpass(constant(0))
        self._end_angle = argpass(constant(2 * pi))
        self._pad_angle = argpass(constant(0))

    def __call__(self, data: Iterable[T], *args) -> list[dict]:
        """
        Generates a pie for the given array of data, returning an array of
        objects representing each datum's arc angles.

        Parameters
        ----------
        data : Iterable[T]
            Data input
        args : Any
            Additional arguments propagated to the pie's accessor functions

        Returns
        -------
        list[dict]
            List of arcs

        Examples
        --------

        For example, given a set of numbers, here is how to compute the
        angles for a pie chart:

        >>> data = [1, 1, 2, 3, 5, 8, 13, 21]
        >>> pie = d3.pie()
        >>> pie(data)

        Each object in the returned array has the following properties:

        - :code:`data` - the input datum; the corresponding element in the input data array.
        - :code:`value` - the numeric value of the arc.
        - :code:`index` - the zero-based sorted index of the arc.
        - :code:`start_angle` - the start angle of the arc.
        - :code:`end_angle` - the end angle of the arc.
        - :code:`pad_angle` - the pad angle of the arc.
        """
        data = list(data)
        n = len(data)

        sum = 0
        index = [None] * n
        arcs = [None] * n
        a0 = self._start_angle(*args)
        da = min(2 * pi, max(-2 * pi, self._end_angle(*args) - a0))
        p = min(abs(da) / n, self._pad_angle(*args))
        pa = p * (-1 if da < 0 else 1)

        for i in range(n):
            d = data[i]
            v = self._value(d, i, data)
            index[i] = i
            arcs[i] = v
            if v > 0:
                sum += v

        if self._sort_values is not None:
            index = sorted(index, key=lambda i: self._sort_values(arcs[i]))
        elif self._sort is not None:
            index = sorted(index, key=lambda i: self._sort(data[i]))

        k = (da - n * pa) / sum if sum else 0
        for i in range(n):
            j = index[i]
            v = arcs[j]
            a1 = a0 + (v * k if v > 0 else 0) + pa
            arcs[j] = asdict(Arc(data[j], i, v, a0, a1, p))
            a0 = a1

        return arcs

    def set_value(self, value: Accessor[T, float] | Number) -> Pie:
        """
        If value is specified, sets the value accessor to the
        specified function or number and returns this pie generator.

        Parameters
        ----------
        value : Accessor[T, float] | Number
            Number or accessor function

        Returns
        -------
        Pie
            Itself
        """
        if callable(value):
            self._value = value
        else:
            self._value = constant(value)
        self._value = argpass(self._value)
        return self

    def set_sort_values(self, sort_values: Callable[[float], float]) -> Pie:
        """
        If sort_values is specified, sets the value comparator to
        the specified function and returns this pie generator.

        Parameters
        ----------
        sort_values : Callable[[float], float]
            Value input

        Returns
        -------
        Pie
            Itself
        """
        self._sort_values = sort_values
        self._sort = None
        return self

    def set_sort(self, sort: Callable[[float], float]) -> Pie:
        """
        If sort is specified, sets the data comparator to
        the specified function and returns this pie generator.

        Parameters
        ----------
        sort : Callable[[float], float]
            Sort function

        Returns
        -------
        Pie
            Itself
        """
        self._sort = sort
        self._sort_values = None
        return self

    def set_start_angle(self, start_angle: Callable[..., float] | Number) -> Pie:
        """
        If start_angle is specified, sets the overall start angle of the
        pie to the specified function or number and returns this pie generator.

        Parameters
        ----------
        start_value : Callable[[..., float]] | Number
            Start angle input

        Returns
        -------
        Pie
            Itself
        """
        if callable(start_angle):
            self._start_angle = start_angle
        else:
            self._start_angle = constant(start_angle)
        self._start_angle = argpass(self._start_angle)
        return self

    def set_end_angle(self, end_angle: Callable[..., float] | Number) -> Pie:
        """
        If end_angle is specified, sets the overall end angle
        of the pie to the specified function or number and
        returns this pie generator.

        Parameters
        ----------
        end_angle : Callable[..., float] | Number
            End angle input

        Returns
        -------
        Pie
            Itself
        """
        if callable(end_angle):
            self._end_angle = end_angle
        else:
            self._end_angle = constant(end_angle)
        self._end_angle = argpass(self._end_angle)
        return self

    def set_pad_angle(self, pad_angle: Callable[..., float] | Number) -> Pie:
        """
        If pad_angle is specified, sets the pad angle to
        the specified function or number and returns this
        pie generator.

        Parameters
        ----------
        pad_angle : Callable[..., float] | Number
            Pad angle input

        Returns
        -------
        Pie
            Itself
        """
        if callable(pad_angle):
            self._pad_angle = pad_angle
        else:
            self._pad_angle = constant(pad_angle)
        self._pad_angle = argpass(self._pad_angle)
        return self

    def get_value(self) -> Accessor[T, float]:
        return self._value

    def get_sort(self) -> Callable[[float], float]:
        return self._sort

    def get_sort_values(self) -> Callable[[float], float]:
        return self._sort_values

    def get_start_angle(self) -> Callable[..., float]:
        return self._start_angle

    def get_end_angle(self) -> Callable[..., float]:
        return self._end_angle

    def get_pad_angle(self) -> Callable[..., float]:
        return self._pad_angle



from __future__ import annotations

from math import sqrt

from ..hierarchy import Node
from .dice import dice
from .slice import slice

PHI = (1 + sqrt(5)) * 0.5


def squarify_ratio(
    ratio: float, parent: Node, x0: float, y0: float, x1: float, y1: float
) -> list[Node]:
    rows = []
    nodes = parent.children
    i0 = 0
    i1 = 0
    n = len(nodes)
    value = parent.value

    while i0 < n:
        dx = x1 - x0
        dy = y1 - y0

        while True:
            sum_value = nodes[i1].value
            i1 += 1
            if sum_value or i1 >= n:
                break

        min_value = max_value = sum_value
        alpha = 0
        if value != 0 and dx != 0 and dy != 0:
            alpha = max(dy / dx, dx / dy) / (value * ratio)
        beta = sum_value * sum_value * alpha
        min_ratio = 0
        if beta != 0:
            min_ratio = max(max_value / beta, beta / min_value)

        while i1 < n:
            node_value = nodes[i1].value
            sum_value += node_value
            if node_value < min_value:
                min_value = node_value
            if node_value > max_value:
                max_value = node_value
            beta = sum_value * sum_value * alpha
            new_ratio = 0
            if beta != 0:
                new_ratio = max(max_value / beta, beta / min_value)
            if new_ratio > min_ratio:
                sum_value -= node_value
                break
            min_ratio = new_ratio
            i1 += 1

        row = Node(None)
        row.value = sum_value
        row.dice = dx < dy
        row.children = nodes[i0:i1]
        rows.append(row)
        if row.dice:
            if value:
                y = y0 + dy * sum_value / value
                dice(row, x0, y0, x1, y)
                y0 = y
            else:
                dice(row, x0, y0, x1, y1)
        else:
            if value:
                x = x0 + dx * sum_value / value
                slice(row, x0, y0, x, y1)
                x0 = x
            else:
                slice(row, x0, y0, x1, y1)
        value -= sum_value
        i0 = i1
    return rows


class Squarify:
    """
    Squarify object

    Parameters
    ----------
    ratio : float
        Ratio value
    """

    def __init__(self, ratio: float):
        self._ratio = ratio

    def __call__(self, parent: Node, x0: float, y0: float, x1: float, y1: float):
        """
        Implements the squarified treemap algorithm by Bruls et al., which
        seeks to produce rectangles of a given aspect ratio.

        Parameters
        ----------
        parent : Node
            Parent node
        x0 : float
            X-coordinate rectangular edge
        y0 : float
            Y-coordinate rectangular edge
        x1 : float
            X-coordinate rectangular edge
        y1 : float
            Y-coordinate rectangular edge
        """
        squarify_ratio(self._ratio, parent, x0, y0, x1, y1)

    def set_ratio(self, ratio: float) -> Squarify:
        """
        Specifies the desired aspect ratio of the generated rectangles. The
        ratio must be specified as a number greater than or equal to one. Note
        that the orientation of the generated rectangles (tall or wide) is not
        implied by the ratio; for example, a ratio of two will attempt to
        produce a mixture of rectangles whose width:height ratio is either
        :code:`2:1` or :code:`1:2`. (However, you can approximately achieve
        this result by generating a square treemap at different dimensions, and
        then stretching the treemap to the desired aspect ratio.) Furthermore,
        the specified ratio is merely a hint to the tiling algorithm; the
        rectangles are not guaranteed to have the specified aspect ratio. If
        not specified, the aspect ratio defaults to the golden ratio,
        :math:`\\phi = (1 + \\sqrt{5}) / 2`, per `Kong et
        al<http://vis.stanford.edu/papers/perception-treemaps>_`.

        Parameters
        ----------
        ratio : float
            Ratio value

        Returns
        -------
        Squarify
            New Squarify object
        """
        return Squarify(ratio if ratio > 1 else 1)


squarify = Squarify(PHI)



import math
from collections.abc import Callable

from ...selection import Selection
from ...types import Number
from .cardinal import BezierTrait
from .common import Curve, isvaluable


class CardinalOpenCurve(Curve, BezierTrait):
    def __init__(self, context, tension):
        self._context = context
        self._k = (1 - tension) / 6
        self._line = math.nan

    def area_start(self):
        self._line = 0

    def area_end(self):
        self._line = math.nan

    def line_start(self):
        self._x0 = math.nan
        self._y0 = math.nan
        self._x1 = math.nan
        self._y1 = math.nan
        self._x2 = math.nan
        self._y2 = math.nan
        self._point = 0

    def line_end(self):
        if isvaluable(self._line) or (self._line != 0 and self._point == 3):
            self._context.close_path()
        self._line = 1 - self._line

    def point(self, x, y):
        if self._point == 0:
            self._point = 1
        elif self._point == 1:
            self._point = 2
        elif self._point == 2:
            self._point = 3
            if isvaluable(self._line):
                self._context.line_to(self._x2, self._y2)
            else:
                self._context.move_to(self._x2, self._y2)
        elif self._point == 3:
            self._point = 4
            self._bezier_curve_to(x, y)
        else:
            self._bezier_curve_to(x, y)
        self._x0 = self._x1
        self._x1 = self._x2
        self._x2 = x
        self._y0 = self._y1
        self._y1 = self._y2
        self._y2 = y


def curve_cardinal_open(
    context_or_tension: Selection | Number,
) -> Callable[[Selection], Curve] | Curve:
    """
    Produces a cubic cardinal spline using the specified control points. Unlike
    curveCardinal, one-sided differences are not used for the first and last
    piece, and thus the curve starts at the second point and ends at the
    penultimate point.
    The default tension is :math:`0`.

    Parameters
    ----------
    context_or_tension : Selection | Number
        Context or tension value in range :math:`[0, 1]` determining the length
        of the tangents. A tension of one yields all zero tangents, equivalent
        to :func:`d3.curve_linear <curve_linear>`

    Returns
    -------
    Callable[[Selection], Curve] | Curve
        Curve object or function which makes a curve object with tension value
        set

    Examples
    --------

    >>> points = [
    ...     [2, 2],
    ...     [6, 8],
    ...     [10, 10],
    ...     [12, 10],
    ...     [14, 4],
    ...     [20, 4],
    ...     [24, 8],
    ...     [29, 6],
    ...     [32, 4],
    ...     [35, 5],
    ...     [38, 2],
    ... ]
    >>> line(points)
    'M6,8C7.333,9.333,9,9.667,10,10C11,10.333,11.333,11,12,10C12.667,9,12.667,5,14,4C15.333,3,18.333,3.333,20,4C21.667,4.667,22.500,7.667,24,8C25.500,8.333,27.667,6.667,29,6C30.333,5.333,31,4.167,32,4C33,3.833,34,5.333,35,5'
    >>> line = d3.line().curve(d3.curve_cardinal_open(0.5))
    >>> line(points)
    'M6,8C6.667,8.667,9.500,9.833,10,10C10.500,10.167,11.667,10.500,12,10C12.333,9.500,13.333,4.500,14,4C14.667,3.500,19.167,3.667,20,4C20.833,4.333,23.250,7.833,24,8C24.750,8.167,28.333,6.333,29,6C29.667,5.667,31.500,4.083,32,4C32.500,3.917,34.500,5.167,35,5'

    **Result**

    .. image:: ../../figures/light_curve_cardinal_open.svg
       :align: center
       :class: only-light

    .. image:: ../../figures/dark_curve_cardinal_open.svg
       :align: center
       :class: only-dark
    """
    if isinstance(context_or_tension, (int, float)):
        tension = context_or_tension

        def local_curve(context):
            return CardinalOpenCurve(context, tension)

        return local_curve
    context = context_or_tension
    return CardinalOpenCurve(context, 0.0)



from math import asin, atan2, cos, degrees, fsum, hypot, nan, radians, sin, sqrt

from ..types import GeoJSON, Point2D
from .common import PolygonStream
from .stream import geo_stream

EPSILON = 1e-6
EPSILON2 = 1e-12


class CentroidStream(PolygonStream):
    def __init__(self):
        self._W0 = 0
        self._W1 = 0
        self._X0 = 0
        self._Y0 = 0
        self._Z0 = 0
        self._X1 = 0
        self._Y1 = 0
        self._Z1 = 0

        self._X2 = []
        self._Y2 = []
        self._Z2 = []

        self._x0 = None
        self._y0 = None
        self._z0 = None

        self._lambda00 = None
        self._phi00 = None

        self._point = self._centroid_point
        self._line_start = self._centroid_line_start
        self._line_end = self._centroid_line_end

    def point(self, x: float, y: float):
        return self._point(x, y)

    def line_start(self):
        return self._line_start()

    def line_end(self):
        return self._line_end()

    def polygon_start(self):
        self._line_start = self._centroid_ring_start
        self._line_end = self._centroid_ring_end

    def polygon_end(self):
        self._line_start = self._centroid_line_start
        self._line_end = self._centroid_line_end

    def sphere(self):
        return

    def _centroid_point(self, lambda_: float, phi: float):
        lambda_ = radians(lambda_)
        phi = radians(phi)
        cos_phi = cos(phi)
        self._centroid_point_cartesian(
            cos_phi * cos(lambda_), cos_phi * sin(lambda_), sin(phi)
        )

    def _centroid_point_cartesian(self, x: float, y: float, z: float):
        self._W0 += 1
        self._X0 += (x - self._X0) / self._W0
        self._Y0 += (y - self._Y0) / self._W0
        self._Z0 += (z - self._Z0) / self._W0

    def _centroid_line_start(self):
        self._point = self._centroid_line_point_first

    def _centroid_line_point_first(self, lambda_: float, phi: float):
        lambda_ = radians(lambda_)
        phi = radians(phi)
        cos_phi = cos(phi)
        self._x0 = cos_phi * cos(lambda_)
        self._y0 = cos_phi * sin(lambda_)
        self._z0 = sin(phi)
        self._point = self._centroid_line_point
        self._centroid_point_cartesian(self._x0, self._y0, self._z0)

    def _centroid_line_point(self, lambda_: float, phi: float):
        lambda_ = radians(lambda_)
        phi = radians(phi)
        cos_phi = cos(phi)
        x = cos_phi * cos(lambda_)
        y = cos_phi * sin(lambda_)
        z = sin(phi)
        w = atan2(
            sqrt(
                (self._y0 * z - self._z0 * y) * (self._y0 * z - self._z0 * y)
                + (self._z0 * x - self._x0 * z) * (self._z0 * x - self._x0 * z)
                + (self._x0 * y - self._y0 * x) * (self._x0 * y - self._y0 * x)
            ),
            self._x0 * x + self._y0 * y + self._z0 * z,
        )

        self._W1 += w
        self._X1 += w * (self._x0 + x)
        self._Y1 += w * (self._y0 + y)
        self._Z1 += w * (self._z0 + z)
        self._x0 = x
        self._y0 = y
        self._z0 = z
        self._centroid_point_cartesian(self._x0, self._y0, self._z0)

    def _centroid_line_end(self):
        self._point = self._centroid_point

    def _centroid_ring_start(self):
        self._point = self._centroid_ring_point_first

    def _centroid_ring_end(self):
        self._centroid_ring_point(self._lambda00, self._phi00)
        self._point = self._centroid_point

    def _centroid_ring_point_first(self, lambda_: float, phi: float):
        self._lambda00 = lambda_
        self._phi00 = phi
        lambda_ = radians(lambda_)
        phi = radians(phi)
        self._point = self._centroid_ring_point
        cos_phi = cos(phi)
        self._x0 = cos_phi * cos(lambda_)
        self._y0 = cos_phi * sin(lambda_)
        self._z0 = sin(phi)
        self._centroid_point_cartesian(self._x0, self._y0, self._z0)

    def _centroid_ring_point(self, lambda_: float, phi: float):
        lambda_ = radians(lambda_)
        phi = radians(phi)
        cos_phi = cos(phi)
        x = cos_phi * cos(lambda_)
        y = cos_phi * sin(lambda_)
        z = sin(phi)
        cx = self._y0 * z - self._z0 * y
        cy = self._z0 * x - self._x0 * z
        cz = self._x0 * y - self._y0 * x
        m = hypot(cx, cy, cz)
        w = asin(m)
        v = -w / m if m != 0 else 0.0

        self._X2.append(v * cx)
        self._Y2.append(v * cy)
        self._Z2.append(v * cz)
        self._W1 += w
        self._X1 += w * (self._x0 + x)
        self._Y1 += w * (self._y0 + y)
        self._Z1 += w * (self._z0 + z)
        self._x0 = x
        self._y0 = y
        self._z0 = z
        self._centroid_point_cartesian(self._x0, self._y0, self._z0)

    def result(self) -> Point2D:
        x = fsum(self._X2)
        y = fsum(self._Y2)
        z = fsum(self._Z2)
        m = hypot(x, y, z)
        if m < EPSILON2:
            x = self._X1
            y = self._Y1
            z = self._Z1
            if self._W1 < EPSILON:
                x = self._X0
                y = self._Y0
                z = self._Z0
            m = hypot(x, y, z)
            if m < EPSILON2:
                return [nan, nan]
        return [degrees(atan2(y, x)), degrees(asin(z / m))]


def geo_centroid(obj: GeoJSON) -> Point2D:
    """
    Returns the spherical centroid of the specified GeoJSON object. This is the
    spherical equivalent of :func:`GeoPath.centroid
    <detroit.geo.path.path.GeoPath.centroid>`.

    Parameters
    ----------
    obj : GeoJSON
        GeoJSON object

    Returns
    -------
    Point2D
        Spherical centroid
    """
    stream = CentroidStream()
    geo_stream(obj, stream)
    return stream.result()



from __future__ import annotations

import math
from collections.abc import Callable
from datetime import datetime, timedelta


class TimeInterval:
    def __call__(self, date: datetime | None = None) -> datetime:
        """
        Applies a *floor* operation on the input. If the input is :code:`None`,
        the input is replaced by the current date.

        Parameters
        ----------
        date : datetime | None
            Date

        Returns
        -------
        datetime
            Date

        Examples
        --------

        >>> from datetime import datetime
        >>> d3.time_day(datetime(2010, 1, 1, 12, 0))
        datetime.datetime(2010, 1, 1, 0, 0)
        """
        date = datetime.now() if date is None else date
        return self.floor(date)

    def interval(self, date: datetime | None = None) -> datetime:
        """
        Same as :code:`__call__`

        Parameters
        ----------
        date : datetime | None
            Date

        Returns
        -------
        datetime
            Date

        Examples
        --------

        >>> from datetime import datetime
        >>> d3.time_day.interval(datetime(2010, 1, 1, 12, 0))
        datetime.datetime(2010, 1, 1, 0, 0)
        """
        return self(date)

    def every(self, step: int) -> TimeInterval:
        """
        Returns a filtered view of this interval representing
        every stepth date. The meaning of step is dependent on this
        interval's parent interval as defined by the field function.

        Parameters
        ----------
        step : int
            Step

        Returns
        -------
        TimeFilter
            Modified class with new :code:`floor` and :code:`offset` methods

        Examples
        --------
        >>> from datetime import datetime
        >>> every = d3.time_day.every(3)
        >>> r = every.range(datetime(2010, 1, 1), datetime(2010, 1, 8))
        >>> for d in r:
        ...     print(d)
        ...
        ...
        2010-01-01 00:00:00
        2010-01-04 00:00:00
        2010-01-07 00:00:00
        """
        step = int(step)
        if math.isinf(step) or not step > 0:
            return None
        if not step > 1:
            return self
        return self.filter(
            (lambda d: self.field(d) % step == 0)
            if self.field is not None
            else (lambda d: self.count(0, d) % step == 0)
        )

    def ceil(self, date: datetime) -> datetime:
        """
        Returns a new date representing the earliest interval
        boundary date after or equal to date.

        Parameters
        ----------
        date : datetime
            Date

        Returns
        -------
        datetime
            Ceiled date

        Examples
        --------

        >>> from datetime import datetime
        >>> d3.time_day.ceil(datetime(2010, 1, 1, 12, 0))
        datetime.datetime(2010, 1, 2, 0, 0)
        """
        return self.floor(self.offset(self.floor(date + timedelta(microseconds=-1)), 1))

    def round(self, date: datetime) -> datetime:
        """
        Returns a new date representing the closest interval
        boundary date to date.

        Parameters
        ----------
        date : datetime
            Date

        Returns
        -------
        datetime
            Rounded date

        Examples
        --------

        >>> from datetime import datetime
        >>> d3.time_day.round(datetime(2010, 1, 1, 12, 0))
        datetime.datetime(2010, 1, 2, 0, 0)
        >>> d3.time_day.round(datetime(2010, 1, 1, 1, 0))
        datetime.datetime(2010, 1, 1, 0, 0)
        """
        d0 = self.interval(date)
        d1 = self.ceil(date)
        return d0 if date - d0 < d1 - date else d1

    def range(self, start: datetime, stop: datetime, step: int = 1) -> list[datetime]:
        """
        Returns an array of dates representing every interval boundary after or
        equal to start (inclusive) and before stop (exclusive).

        Parameters
        ----------
        start : datetime
            Start date
        stop : datetime
            Stop date
        step : int
            Step

        Returns
        -------
        list[datetime]
            Range of dates

        Examples
        --------
        >>> from datetime import datetime
        >>> r = d3.time_day.range(datetime(2010, 1, 1), datetime(2010, 1, 5), step=1)
        >>> for d in r:
        ...     print(d)
        ...
        ...
        2010-01-01 00:00:00
        2010-01-02 00:00:00
        2010-01-03 00:00:00
        2010-01-04 00:00:00
        >>> r = d3.time_day.range(datetime(2010, 1, 1), datetime(2010, 1, 5), step=2)
        >>> for d in r:
        ...     print(d)
        ...
        ...
        2010-01-01 00:00:00
        2010-01-03 00:00:00
        """
        range_list = []
        start = self.ceil(start)
        if not (start < stop) or not (step > 0):
            return range_list

        previous = None
        while previous is None or (previous < start and start < stop):
            previous = start
            range_list.append(previous)
            start = self.floor(self.offset(start, step))

        return range_list

    @classmethod
    def filter(cls, test: Callable[[datetime], bool]) -> TimeInterval:
        """
        Returns a new interval that is a filtered subset of this interval using
        the specified test function.

        Parameters
        ----------
        test : Callable[[datetime], bool]
            Function which returns :code:`True` if and only if the specified
            date should be considered part of the interval

        Returns
        -------
        TimeFilter
            Modified class with new :code:`floor` and :code`offset` function.
        """

        class TimeFilter(cls):
            """
            Modified class which calls several times :code:`floor`
            and :code:`offset` in order to achieve :code:`every(step)`
            function.
            """

            def floor(self, date: datetime):
                """
                Returns a new date representing the latest interval
                boundary date before or equal to date.

                Parameters
                ----------
                date : datetime
                    Date

                Returns
                -------
                datetime
                    Floored date
                """
                date = cls.floor(self, date)
                while not test(date):
                    date = cls.floor(self, date - timedelta(microseconds=1))
                return date

            def offset(self, date: datetime, step: int):
                """
                Returns a new date equal to date plus step intervals.

                Parameters
                ----------
                date : datetime
                    Date
                step : int
                    Step to add or substitute

                Returns
                -------
                datetime
                    Date
                """
                if step < 0:
                    step += 1
                    while step <= 0:
                        date = cls.offset(self, date, -1)
                        while not test(date):
                            date = cls.offset(self, date, -1)
                        step += 1
                else:
                    step -= 1
                    while step >= 0:
                        date = cls.offset(self, date, 1)
                        while not test(date):
                            date = cls.offset(self, date, 1)
                        step -= 1
                return date

        return TimeFilter()



from __future__ import annotations

import math
from collections.abc import Callable
from datetime import datetime
from typing import overload

from ..array import ticks
from ..format import format_specifier, locale_format
from ..types import Number
from .continuous import Transformer, copy
from .init import init_range
from .nice import nice


def transform_log(x: datetime | float) -> float:
    if isinstance(x, datetime):
        x = x.timestamp()
    return math.log(x)


def transform_exp(x: datetime | float) -> float:
    if isinstance(x, datetime):
        x = x.timestamp()
    return math.exp(x)


def transform_logn(x: datetime | float) -> float:
    if isinstance(x, datetime):
        x = x.timestamp()
    return -math.log(-x)


def transform_expn(x: datetime | float) -> float:
    if isinstance(x, datetime):
        x = x.timestamp()
    return -math.exp(-x)


def pow10(x: datetime | float) -> float:
    if isinstance(x, datetime):
        x = x.timestamp()
    return 10**x if math.isfinite(x) else 0 if x < 0 else x


def powp(base: Number) -> Callable[[float], float]:
    if base == 10:
        return pow10
    elif base == math.e:
        return math.exp
    else:
        return lambda x: math.pow(base, x)


def logp(base: Number) -> Callable[[float], float]:
    if base == math.e:
        return math.log
    elif base == 10:
        return math.log10
    elif base == 2:
        return math.log2
    else:
        return lambda x: math.log(x, base)


def reflect(f: Callable[[float], float]) -> Callable[[float], float]:
    def local_reflect(x):
        return -f(-x)

    return local_reflect


class LogBase:
    """
    Logarithmic ("log") scales are like linear scales except that
    a logarithmic transform is applied to the input domain value
    before the output range value is computed. The mapping to the
    range value y can be expressed as a function of the domain value x:
    :math:`y = m \\cdot \\log(x) + b`.
    """

    def __init__(self):
        self._base = 10
        self._logs = None
        self._pows = None

    def set_base(self, base: Number) -> LogBase:
        """
        Sets the scale's base value

        Parameters
        ----------
        base : Number
            Base value

        Returns
        -------
        LogBase
            Itself
        """
        self._base = float(base)
        return self._log_rescale()

    def get_base(self) -> Number:
        return self._base

    def ticks(self, count: int | None = None) -> LogBase:
        """
        Like :code:`ScaleLinear.ticks`, but customized for a log scale.

        Parameters
        ----------
        count : int | None
            Count. If specified, the scale may return more or fewer
            values depending on the domain

        Returns
        -------
        LogBase
            Itself
        """
        d = self.get_domain()
        u = d[0]
        v = d[-1]
        r = v < u

        if r:
            u, v = v, u

        i = self._logs(u)
        j = self._logs(v)
        k = None
        t = None
        n = count if count is not None else 10
        z = []

        if not (self._base % 1) and j - i < n:
            i = math.floor(i)
            j = math.ceil(j)
            if u > 0:
                for i in range(i, j + 1):
                    for k in range(1, int(self._base)):
                        t = k / self._pows(-i) if i < 0 else k * self._pows(i)
                        if t < u:
                            continue
                        if t > v:
                            break
                        z.append(t)
            else:
                for i in range(i, j + 1):
                    for k in range(int(self._base) - 1, 0, -1):
                        t = k / self._pows(-i) if i > 0 else k * self._pows(i)
                        if t < u:
                            continue
                        if t > v:
                            break
                        z.append(t)
            if len(z) * 2 < n:
                z = ticks(u, v, n)
        else:
            z = list(map(self._pows, ticks(i, j, min(j - i, n))))
        return z[::-1] if r else z

    def tick_format(
        self, count: int | None = None, specifier: str | None = None
    ) -> LogBase:
        """
        Like :code:`ScaleLinear.tick_format`, but customized for a log scale.
        The specified count typically has the same value as the count
        that is used to generate the tick values.

        Parameters
        ----------
        count : int | None
            Count. It should have the same value as the count
            that is used to generate the tick values.
        specifier : str | None
            Specifier


        Returns
        -------
        LogBase
            Itself
        """
        if count is None:
            count = 10
        if specifier is None:
            specifier = "s" if self._base == 10 else ","
        if not callable(specifier):
            specifier_obj = format_specifier(specifier)
            if (
                not (self._base % 1)
                and specifier_obj is not None
                and specifier_obj.precision is None
            ):
                specifier_obj.trim = True
            specifier = locale_format(specifier_obj or specifier)
        if math.isinf(count):
            return specifier
        k = max(1, self._base * count / len(self.ticks()))

        def f(d):
            i = d / self._pows(round(self._logs(d)))
            if i * self._base < self._base - 0.5:
                i *= self._base
            return specifier(d) if i <= k else ""

        return f

    def nice(self) -> LogBase:
        """
        Like :code:`ScaleLinear.nice`, except extends the domain
        to integer powers of base.

        Returns
        -------
        LogBase
            Itself
        """

        class Interval:
            @staticmethod
            def floor(x):
                if x == 0:
                    return 0
                return self._pows(math.floor(self._logs(x)))

            @staticmethod
            def ceil(x):
                if x == 0:
                    return 0
                return self._pows(math.ceil(self._logs(x)))

        return self.set_domain(nice(self.get_domain(), Interval))


class ScaleLog(Transformer[float], LogBase):
    """
    Logarithmic ("log") scales are like linear scales except that a logarithmic
    transform is applied to the input domain value before the output range value
    is computed. The mapping to the range value y can be expressed as a function
    of the domain value x: :math:`y = m \\log(x) + b`.
    """

    def __init__(self):
        Transformer.__init__(self, transform_log, transform_exp)
        LogBase.__init__(self)

    def _log_rescale(self):
        self._logs = logp(self._base)
        self._pows = powp(self._base)
        d = self.get_domain()[0]
        if isinstance(d, datetime):
            d = d.timestamp()
        if d < 0:
            self._logs = reflect(self._logs)
            self._pows = reflect(self._pows)
            self._transform = transform_logn
            self._untransform = transform_expn
            self._rescale()
        else:
            self._transform = transform_log
            self._untransform = transform_exp
            self._rescale()
        return self

    def set_domain(self, domain: list[float]) -> ScaleLog:
        """
        Sets the scale's domain to the specified array of values.

        Parameters
        ----------
        domain : list[float]
            Domain

        Returns
        -------
        ScaleLog
            Itself
        """
        super().set_domain(domain)
        return self._log_rescale()

    def copy(self):
        return copy(self, ScaleLog()).set_base(self.get_base())


@overload
def scale_log() -> ScaleLog: ...


@overload
def scale_log(range_vals: list[Number]) -> ScaleLog: ...


@overload
def scale_log(domain: list[Number], range_vals: list[Number]) -> ScaleLog: ...


def scale_log(*args):
    """
    Builds a new log scale with the specified domain and range,
    the base 10, the default interpolator and clamping disabled.

    Parameters
    ----------
    domain : list[Number]
        Array of numbers
    range_vals : list[Number]
        Array of values

    Returns
    -------
    ScaleLog
        Scale object

    Examples
    --------

    >>> scale = d3.scale_log([1, 10], [0, 960])
    >>> steps = 10
    >>> for x in range(steps + 1):
    ...     x = x / steps
    ...     x = 10 ** x
    ...     print(x, scale(x))
    ...
    ...
    1.0 0.0
    1.2589254117941673 96.0
    1.5848931924611136 192.0
    1.9952623149688795 287.99999999999994
    2.51188643150958 383.99999999999994
    3.1622776601683795 480.0
    3.9810717055349722 576.0
    5.011872336272722 671.9999999999999
    6.309573444801933 767.9999999999999
    7.943282347242816 864.0
    10.0 960.0
    """
    scale = ScaleLog().set_domain([1, 10])
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)



from __future__ import annotations

from collections.abc import Iterator
from typing import Any, Generic

from ..types import T


class Serie(Generic[T]):
    """
    Point of a :code:`Series`

    Parameters
    ----------
    values : list[float]
        Point coordinates
    data : T
        Data value associated to this point
    """

    def __init__(self, values: list[float], data: T):
        self._values = values
        self.data = data

    def __getitem__(self, index: int) -> float:
        """
        Returns the value of a specific index

        Parameters
        ----------
        index : int
            Index value

        Returns
        -------
        float
            Value
        """
        return self._values[index]

    def __setitem__(self, index: int, value: float):
        """
        Sets the value at a specific index

        Parameters
        ----------
        index : int
            Index value
        value : float
            Value
        """
        self._values[index] = value

    def __len__(self) -> int:
        """
        Returns the length of the serie

        Returns
        -------
        int
            Length of serie
        """
        return len(self._values)

    def __eq__(self, serie: Serie | Any) -> bool:
        """
        Checks if another series value is the same as the current values.

        Parameters
        ----------
        serie : Serie[T] | Any
            Other serie

        Returns
        -------
        bool
            True if two series are the same
        """
        if not isinstance(serie, Serie):
            return False
        return self._values == serie._values and self.data == serie.data

    def __str__(self) -> str:
        """
        Formats the class to a string.

        Returns
        -------
        str
           String representing the current class
        """
        return f"Serie({self._values}, data={self.data})"

    def __repr__(self):
        return str(self)


class Series(Generic[T]):
    """
    List of :code:`Serie` associated to a :code:`key` and an :code:`index`

    Parameters
    ----------
    series: list[Serie[T]] | None
        List of series
    """

    def __init__(self, series: list[Serie[T]] | None = None):
        self._series = series or []
        self.key = None
        self.index = None

    def __getitem__(self, index: int) -> Serie[T]:
        """
        Gets a serie given an index

        Parameters
        ----------
        index : int
            Index value

        Returns
        -------
        Serie[T]
            Serie value
        """
        return self._series[index]

    def __setitem__(self, index: int, serie: Serie[T]):
        """
        Sets a serie at a specific index

        Parameters
        ----------
        index : int
            Index value
        serie : Serie[T]
            Serie value
        """
        self._series[index] = serie

    def __len__(self) -> int:
        """
        Returns the length of the serie

        Returns
        -------
        int
            Length
        """
        return len(self._series)

    def __eq__(self, series: Series | Any) -> bool:
        """
        Checks if another series value is the same as the current values.

        Parameters
        ----------
        series : Series[T] | Any
            Other series

        Returns
        -------
        bool
            True if two series are the same
        """
        if not isinstance(series, Series):
            return False
        return (
            self._series == series._series
            and self.key == series.key
            and self.index == series.index
        )

    def __iter__(self) -> Iterator[Serie[T]]:
        """
        Returns an iterator of :code:`Serie`.

        Returns
        -------
        Iterator[Serie[T]]
            Iterator
        """
        return iter(self._series)

    def append(self, serie: Serie[T]):
        """
        Appends a serie to the current list of series.

        Parameters
        ----------
        serie : Serie[T]
            Serie value
        """
        self._series.append(serie)

    def __str__(self) -> str:
        """
        Formats the class to a string.

        Returns
        -------
        str
           String representing the current class
        """
        return f"Series({self._series}, key={self.key}, index={self.index})"

    def __repr__(self):
        return str(self)



from .antimeridian import geo_clip_antimeridian
from .circle import geo_clip_circle
from .rectangle import geo_clip_rectangle

__all__ = [
    "geo_clip_antimeridian",
    "geo_clip_circle",
    "geo_clip_rectangle",
]



from math import nan, radians

from ..types import GeoJSON, Point2D
from .distance import geo_distance
from .polygon_contains import polygon_contains

EPSILON2 = 1e-12


class ContainsObjectType:
    def Feature(obj: GeoJSON, point: Point2D) -> bool:
        return contains_geometry(obj["geometry"], point)

    def FeatureCollection(obj: GeoJSON, point: Point2D) -> bool:
        for feature in obj["features"]:
            if contains_geometry(feature["geometry"], point):
                return True
        return False


class ContainsGeometryType:
    def Sphere(obj: GeoJSON, point: Point2D) -> bool:
        return True

    def Point(obj: GeoJSON, point: Point2D) -> bool:
        return contains_point(obj["coordinates"], point)

    def MultiPoint(obj: GeoJSON, point: Point2D) -> bool:
        for coordinates in obj["coordinates"]:
            if contains_point(coordinates, point):
                return True
        return False

    def LineString(obj: GeoJSON, point: Point2D) -> bool:
        return contains_line(obj["coordinates"], point)

    def MultiLineString(obj: GeoJSON, point: Point2D) -> bool:
        for coordinates in obj["coordinates"]:
            if contains_line(coordinates, point):
                return True
        return False

    def Polygon(obj: GeoJSON, point: Point2D) -> bool:
        return contains_polygon(obj["coordinates"], point)

    def MultiPolygon(obj: GeoJSON, point: Point2D) -> bool:
        for coordinates in obj["coordinates"]:
            if contains_polygon(coordinates, point):
                return True
        return False

    def GeometryCollection(obj: GeoJSON, point: Point2D) -> bool:
        for geometry in obj["geometries"]:
            if contains_geometry(geometry, point):
                return True
        return False


def contains_geometry(geometry: GeoJSON, point: Point2D) -> bool:
    if geometry and hasattr(ContainsGeometryType, geometry["type"]):
        return getattr(ContainsGeometryType, geometry["type"])(geometry, point)
    return False


def contains_point(coordinates: GeoJSON, point: Point2D) -> bool:
    return geo_distance(coordinates, point) == 0


def contains_line(coordinates: list[Point2D], point: Point2D) -> bool:
    ao = nan
    for i in range(len(coordinates)):
        bo = geo_distance(coordinates[i], point)
        if bo == 0:
            return True
        if i > 0:
            ab = geo_distance(coordinates[i], coordinates[i - 1])
            if (
                ab > 0
                and ao <= ab
                and bo <= ab
                and (ao + bo - ab) * (1 - pow((ao - bo) / ab, 2)) < EPSILON2 * ab
            ):
                return True
        ao = bo
    return False


def contains_polygon(coordinates: list[list[Point2D]], point: Point2D) -> bool:
    return not (
        not (
            polygon_contains(list(map(ring_radians, coordinates)), point_radians(point))
        )
    )


def ring_radians(ring: list[Point2D]) -> list[Point2D]:
    ring = [point_radians(p) for p in ring]
    ring.pop()
    return ring


def point_radians(point: Point2D):
    return [radians(point[0]), radians(point[1])]


def geo_contains(obj: GeoJSON, point: Point2D) -> bool:
    """
    Returns :code:`True` if and only if the specified GeoJSON object contains
    the specified point, or :code:`False` if the object does not contain the
    point. The point must be specified as a two-element array
    :code:`[longitude, latitude]` in degrees. For Point and MultiPoint
    geometries, an exact test is used; for a Sphere, true is always returned;
    for other geometries, an epsilon threshold is applied.

    Parameters
    ----------
    obj : GeoJSON
        GeoJSON object
    point : Point2D
        2D point

    Returns
    -------
    bool
       :code:`True` if the GeoJSON object contains the point else :code:`False`
    """
    if obj and hasattr(ContainsObjectType, obj["type"]):
        return getattr(ContainsObjectType, obj["type"])(obj, point)
    else:
        return contains_geometry(obj, point)



from .argpass import argpass
from .bin import bin
from .blur import blur, blur2, blur_image
from .extent import extent
from .group import group, groups, index, indexes, rollup, rollups
from .nice import nice
from .set_operations import (
    difference,
    disjoint,
    intersection,
    subset,
    superset,
    union,
)
from .threshold import threshold_freedman_diaconis, threshold_scott, threshold_sturges
from .ticks import tick_increment, tick_step, ticks

__all__ = [
    "argpass",
    "bin",
    "blur",
    "blur2",
    "blur_image",
    "difference",
    "disjoint",
    "extent",
    "group",
    "groups",
    "index",
    "indexes",
    "intersection",
    "nice",
    "rollup",
    "rollups",
    "subset",
    "superset",
    "threshold_freedman_diaconis",
    "threshold_scott",
    "threshold_sturges",
    "tick_increment",
    "tick_step",
    "ticks",
    "union",
]



from ..colors import colors
from ..ramp import ramp

scheme = [
    "fc8d59ffffbf91bfdb",
    "d7191cfdae61abd9e92c7bb6",
    "d7191cfdae61ffffbfabd9e92c7bb6",
    "d73027fc8d59fee090e0f3f891bfdb4575b4",
    "d73027fc8d59fee090ffffbfe0f3f891bfdb4575b4",
    "d73027f46d43fdae61fee090e0f3f8abd9e974add14575b4",
    "d73027f46d43fdae61fee090ffffbfe0f3f8abd9e974add14575b4",
    "a50026d73027f46d43fdae61fee090e0f3f8abd9e974add14575b4313695",
    "a50026d73027f46d43fdae61fee090ffffbfe0f3f8abd9e974add14575b4313695",
]
SCHEME_RDYLBU = list(map(colors, scheme))
interpolate_rdylbu = ramp(SCHEME_RDYLBU)



from ..colors import colors
from ..ramp import ramp

scheme = [
    "deebf79ecae13182bd",
    "eff3ffbdd7e76baed62171b5",
    "eff3ffbdd7e76baed63182bd08519c",
    "eff3ffc6dbef9ecae16baed63182bd08519c",
    "eff3ffc6dbef9ecae16baed64292c62171b5084594",
    "f7fbffdeebf7c6dbef9ecae16baed64292c62171b5084594",
    "f7fbffdeebf7c6dbef9ecae16baed64292c62171b508519c08306b",
]
SCHEME_BLUES = list(map(colors, scheme))
interpolate_blues = ramp(SCHEME_BLUES)



from math import atan2, cos, fsum, radians, sin, sqrt

from ..array import argpass
from ..types import GeoJSON
from .common import PolygonStream
from .stream import geo_stream


@argpass
def noop():
    return


class LengthStream(PolygonStream):
    def __init__(self):
        self._point = noop
        self._line_start = self._length_line_start
        self._line_end = noop

        self._length_sum = []
        self._lambda0 = None
        self._cos_phi0 = None
        self._sin_phi0 = None

    def point(self, x: float, y: float):
        return self._point(x, y)

    def line_start(self):
        return self._line_start()

    def line_end(self):
        return self._line_end()

    def polygon_start(self):
        return

    def polygon_end(self):
        return

    def sphere(self):
        return

    def _length_line_start(self):
        self._point = self._length_point_first
        self._line_end = self._length_line_end

    def _length_line_end(self):
        self._point = noop
        self._line_end = noop

    def _length_point_first(self, lambda_: float, phi: float):
        lambda_ = radians(lambda_)
        phi = radians(phi)
        self._lambda0 = lambda_
        self._sin_phi0 = sin(phi)
        self._cos_phi0 = cos(phi)
        self._point = self._length_point

    def _length_point(self, lambda_: float, phi: float):
        lambda_ = radians(lambda_)
        phi = radians(phi)

        sin_phi = sin(phi)
        cos_phi = cos(phi)
        delta = abs(lambda_ - self._lambda0)
        cos_delta = cos(delta)
        sin_delta = sin(delta)
        x = cos_phi * sin_delta
        y = self._cos_phi0 * sin_phi - self._sin_phi0 * cos_phi * cos_delta
        z = self._sin_phi0 * sin_phi + self._cos_phi0 * cos_phi * cos_delta
        self._length_sum.append(atan2(sqrt(x * x + y * y), z))

        self._lambda0 = lambda_
        self._cos_phi0 = cos_phi
        self._sin_phi0 = sin_phi

    def result(self) -> float:
        return fsum(self._length_sum)


def geo_length(obj: GeoJSON) -> float:
    """
    Returns the great-arc length of the specified GeoJSON object in radians.
    For polygons, returns the perimeter of the exterior ring plus that of any
    interior rings. This is the spherical equivalent of :func:`GeoPath.measure
    <detroit.geo.path.path.GeoPath.measure>`.

    Parameters
    ----------
    obj : GeoJSON
        GeoJSON object

    Returns
    -------
    float
        Great-arc length
    """
    stream = LengthStream()
    geo_stream(obj, stream)
    return stream.result()



from collections.abc import Callable
from math import cos, sin

from ...selection import Selection
from .common import Curve
from .linear import curve_linear


class RadialCurve(Curve):
    def __init__(self, curve: Curve):
        self._curve = curve

    def area_start(self):
        self._curve.area_start()

    def area_end(self):
        self._curve.area_end()

    def line_start(self):
        self._curve.line_start()

    def line_end(self):
        self._curve.line_end()

    def point(self, a: float, r: float):
        self._curve.point(r * sin(a), -r * cos(a))


def curve_radial(
    curve: Callable[[Selection], Curve],
) -> Callable[[Selection], Curve]:
    """
    Converts a curve function into a radial curve function.

    Parameters
    ----------
    curve : Callable[[Selection], Curve]
        Curve function

    Returns
    -------
    Callable[[Selection], Curve]
        Radial curve function
    """

    def radial(context: Selection) -> Curve:
        if isinstance(curve(None), RadialCurve):
            return curve(context)
        return RadialCurve(curve(context))

    return radial


def curve_radial_linear(context: Selection) -> Curve:
    """
    Produces a polyline through the specified polar points.

    Parameters
    ----------
    context : Selection
        Context

    Returns
    -------
    Curve
        Radial curve
    """
    return curve_radial(curve_linear)(context)



from math import atan, atan2, cos, log, pi, pow, sin, sqrt, tan

from ...types import Point2D
from ..common import Projection, RawProjection
from .conic import conic_projection
from .mercator import MercatorRaw

EPSILON = 1e-6
HALF_PI = pi * 0.5


def tany(y):
    return tan((HALF_PI + y) * 0.5)


def sign(x):
    return -1 if x < 0 else 1


class ConicConformalRaw:
    def __init__(self, n: float, f: float):
        self._n = n
        self._f = f

    def __call__(self, x: float, y: float) -> Point2D:
        if self._f > 0.0:
            if y < -HALF_PI + EPSILON:
                y = -HALF_PI + EPSILON
        else:
            if y > HALF_PI - EPSILON:
                y = HALF_PI - EPSILON
        r = self._f / pow(tan((HALF_PI + y) * 0.5), self._n)
        return [r * sin(self._n * x), self._f - r * cos(self._n * x)]

    def invert(self, x: float, y: float) -> Point2D:
        fy = self._f - y
        r = sign(self._n) * sqrt(x * x + fy * fy)
        l_ = atan2(x, abs(fy)) * sign(fy)
        if fy * self._n < 0:
            l_ -= pi * sign(x) * sign(fy)
        return [l_ / self._n, 2 * atan(pow(self._f / r, 1 / self._n)) - HALF_PI]


def conic_conformal_raw(y0: float, y1: float) -> RawProjection:
    cy0 = cos(y0)
    n = sin(y0) if y0 == y1 else log(cy0 / cos(y1)) / log(tany(y1) / tany(y0))
    f = cy0 * pow(tany(y0), n) / n
    return MercatorRaw() if n == 0.0 else ConicConformalRaw(n, f)


def geo_conic_conformal() -> Projection:
    """
    the conic conformal projection. the parallels default to [30°, 30°]
    resulting in flat top.

    returns
    -------
    projection
        projection object
    """
    return conic_projection(conic_conformal_raw).scale(109.5).parallels([30, 30])



from collections.abc import Callable
from itertools import chain
from math import pi

from ...types import T
from ..common import LineStream, PolygonStream, Stream
from ..polygon_contains import polygon_contains
from .buffer import ClipBuffer
from .rejoin import Intersection, clip_rejoin

EPSILON = 1e-6
half_pi = 0.5 * pi


def valid_segment(segment: list[T]) -> bool:
    return len(segment) > 1


def compare_intersection(a: Intersection, b: Intersection) -> float:
    a = a.x
    b = b.x
    ra = a[1] - half_pi - EPSILON if a[0] < 0 else half_pi - a[1]
    rb = b[1] - half_pi - EPSILON if b[0] < 0 else half_pi - b[1]
    return ra - rb


class Clip(PolygonStream):
    def __init__(
        self,
        point_visible: Callable[[float, float], bool],
        clip_line: Callable[[Stream], LineStream],
        interpolate: Callable[[float | None, float | None, float, LineStream], None],
        start: tuple[float, float],
        sink: PolygonStream,
    ):
        self._line = clip_line(sink)
        self._ring_buffer = ClipBuffer()
        self._ring_sink = clip_line(self._ring_buffer)
        self._polygon_started = False
        self._polygon = None
        self._segments = None
        self._ring = None

        self._start = start
        self._sink = sink
        self._point_visible = point_visible
        self._interpolate = interpolate
        self._clip_line = clip_line

        self._line_start = self._line_start_default
        self._line_end = self._line_end_default
        self._point = self._point_default

    def line_start(self):
        return self._line_start()

    def line_end(self):
        return self._line_end()

    def polygon_start(self):
        self._point = self._point_ring
        self._line_start = self._ring_start
        self._line_end = self._ring_end
        self._segments = []
        self._polygon = []

    def polygon_end(self):
        self._point = self._point_default
        self._line_start = self._line_start_default
        self._line_end = self._line_end_default
        self._segments = list(chain.from_iterable(self._segments))
        start_inside = polygon_contains(self._polygon, self._start)
        if len(self._segments) > 0:
            if not self._polygon_started:
                self._sink.polygon_start()
                self._polygon_started = True
            clip_rejoin(
                self._segments,
                compare_intersection,
                start_inside,
                self._interpolate,
                self._sink,
            )
        elif start_inside:
            if not self._polygon_started:
                self._sink.polygon_start()
                self._polygon_started = True
            self._sink.line_start()
            self._interpolate(None, None, 1, self._sink)
            self._sink.line_end()
        if self._polygon_started:
            self._sink.polygon_end()
            self._polygon_started = False
        self._segments = None
        self._polygon = None

    def sphere(self):
        self._sink.polygon_start()
        self._sink.line_start()
        self._interpolate(None, None, 1, self._sink)
        self._sink.line_end()
        self._sink.polygon_end()

    def point(self, lambda_: float, phi: float):
        return self._point(lambda_, phi)

    def _point_default(self, lambda_: float, phi: float):
        if self._point_visible(lambda_, phi):
            self._sink.point(lambda_, phi)

    def _point_line(self, lambda_: float, phi: float):
        self._line.point(lambda_, phi)

    def _line_start_default(self):
        self._point = self._point_line
        self._line.line_start()

    def _line_end_default(self):
        self._point = self._point_default
        self._line.line_end()

    def _point_ring(self, lambda_: float, phi: float):
        self._ring.append([lambda_, phi])
        self._ring_sink.point(lambda_, phi)

    def _ring_start(self):
        self._ring_sink.line_start()
        self._ring = []

    def _ring_end(self):
        self._point_ring(self._ring[0][0], self._ring[0][1])
        self._ring_sink.line_end()

        clean = self._ring_sink.clean()
        ring_segments = self._ring_buffer.result()
        n = len(ring_segments)

        self._ring.pop()
        self._polygon.append(self._ring)
        self._ring = None

        if n == 0:
            return

        if clean & 1:
            segment = ring_segments[0]
            m = len(segment) - 1
            if m > 0:
                if not self._polygon_started:
                    self._sink.polygon_start()
                    self._polygon_started = True
                self._sink.line_start()
                for point in segment[:-1]:
                    self._sink.point(point[0], point[1])
                self._sink.line_end()
            return

        if n > 1 and clean & 2:
            ring_segments.append(ring_segments.pop() + ring_segments.pop(0))
        self._segments.append(list(filter(valid_segment, ring_segments)))

    def __str__(self):
        return (
            f"Clip(point_visible={self._point_visible.__name__}, "
            f"clip_line={self._clip_line.__name__}, "
            f"interpolate={self._interpolate.__name__}, start={self._start}, "
            f"sink={self._sink})"
        )


class ClipWrapper:
    def __init__(
        self,
        point_visible: Callable[[float, float], bool],
        clip_line: Callable[[Stream], LineStream],
        interpolate: Callable[[float | None, float | None, float, LineStream], None],
        start: tuple[float, float],
    ):
        self._point_visible = point_visible
        self._clip_line = clip_line
        self._interpolate = interpolate
        self._start = start

    def __call__(self, sink: PolygonStream) -> Clip:
        return Clip(
            self._point_visible, self._clip_line, self._interpolate, self._start, sink
        )


def clip(
    point_visible: Callable[[float, float], bool],
    clip_line: Callable[[Stream], LineStream],
    interpolate: Callable[[float | None, float | None, float, LineStream], None],
    start: tuple[float, float],
) -> Callable[[PolygonStream], Clip]:
    return ClipWrapper(point_visible, clip_line, interpolate, start)



def format_trim(s):
    n = len(s)
    i0 = -1
    i1 = 0
    for i in range(1, n):
        c = s[i]
        if c == ".":
            i0 = i1 = i
        elif c == "0":
            if i0 == 0:
                i0 = i
            i1 = i
        else:
            if not c.isdigit():
                break
            if i0 > 0:
                i0 = 0
    return s[:i0] + s[i1 + 1 :] if i0 > 0 else s



import re
from collections.abc import Callable

from .number import interpolate_number

reA = re.compile(r"[-+]?(?:\d+\.?\d*|\.?\d+)(?:[eE][-+]?\d+)?")
reB = re.compile(reA.pattern)


def zero(b):
    def f(*args):
        return b

    return f


def one(b):
    def f(t):
        return str(b(t))

    return f


def interpolate_string(a: str, b: str) -> Callable[[float], str]:
    """
    Returns an interpolator between the two strings a and b.

    Parameters
    ----------
    a : str
        String a
    b : str
        String b

    Returns
    -------
    Callable[[float], str]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_string("10.", "20.")
    >>> interpolator(0)
    '10.0'
    >>> interpolator(1)
    '20.0'
    >>> interpolator(0.5)
    '15.0'
    """
    a, b = str(a), str(b)

    bi = 0
    s = []
    q = []

    for am in reA.finditer(a):
        bm = reB.search(b, bi)
        if bm:
            if bm.start() > bi:
                s.append(b[bi : bm.start()])

            if am.group() == bm.group():
                s.append(bm.group())
            else:
                s.append(None)
                q.append(
                    {
                        "i": len(s) - 1,
                        "x": interpolate_number(float(am.group()), float(bm.group())),
                    }
                )

            bi = bm.end()

    if bi < len(b):
        s.append(b[bi:])

    if len(s) < 2:
        return one(q[0]["x"]) if q else zero(b)

    def interpolator(t):
        for o in q:
            s[o["i"]] = str(o["x"](t)).removesuffix(".0")
        return "".join(s)

    return interpolator



from collections.abc import Callable
from datetime import datetime
from typing import Any, Generic, Protocol, TypeAlias, TypedDict, TypeVar, overload

from lxml import etree

T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")
TScaler = TypeVar("Itself", bound="Scaler")
TContinuousScaler = TypeVar("Itself", bound="ContinuousScaler")
TSequentialScaler = TypeVar("Itself", bound="SequentialScaler")
Number: TypeAlias = int | float
GenValue: TypeAlias = datetime | str | int | float
Point2D: TypeAlias = tuple[float, float]
Point3D: TypeAlias = tuple[float, float, float]
Vec2D: TypeAlias = tuple[float, float]
Vec3D: TypeAlias = tuple[float, float, float]
GeoJSON: TypeAlias = dict[str, Any]

# Type definition for :code:`Formatter`: a function which takes a string to be
# formatted and returns the formatted value.
Formatter: TypeAlias = Callable[[str], T]

EtreeFunction: TypeAlias = Callable[[etree.Element, U, int, list[etree.Element]], V]


class Interval(Protocol):
    """
    Protocol class which represents Interval class
    """

    @staticmethod
    def floor(x: float) -> float:
        """
        Apply floor procedure on x value

        Parameters
        ----------
        x : float
            Input

        Returns
        -------
        float
            Output
        """
        ...

    @staticmethod
    def ceil(x: float) -> float:
        """
        Apply ceil procedure on x value

        Parameters
        ----------
        x : float
            Input

        Returns
        -------
        float
            Output
        """
        ...


class Scaler(Protocol[U, V]):
    """
    Protocol class which represents Scaler class
    """

    def __call__(self, x: U) -> V:
        """
        Scales a value.

        Parameters
        ----------
        x : U
            Value to transform

        Returns
        -------
        V
            Scaled value
        """
        ...

    def get_domain(self) -> list[U]:
        """
        Returns the domain values.

        Returns
        -------
        list[U]
            Domain values
        """
        ...

    def get_range(self) -> list[V]:
        """
        Returns the range values.

        Returns
        -------
        list[V]
            Range values
        """
        ...

    def set_domain(self, domain: list[U]) -> TScaler:
        """
        Sets the domain and returns the scaler updated.

        Parameters
        ----------
        domain : list[U]
            Domain values to set

        Returns
        -------
        Scaler
            Itself
        """
        ...

    def set_range(self, range_vals: list[V]) -> TScaler:
        """
        Sets the range and returns the scaler updated.

        Parameters
        ----------
        range_vals : list[V]
            Range values to set

        Returns
        -------
        Scaler
            Itself
        """
        ...


class ContinuousScaler(Scaler[U, V], Generic[U, V]):
    """
    Protocol class which represents Continuous Scaler class
    """

    def invert(self, y: V) -> U:
        """
        Makes the opposite operation of :code:`__call__`.

        Parameters
        ----------
        y : V
            Scaled value

        Returns
        -------
        U
            Original value
        """
        ...

    def get_interpolate(self) -> Callable[[V], V]:
        """
        Gets the interpolator function.

        Returns
        -------
        Callable[[V], V]
            Interpolate function
        """
        ...

    def get_clamp(self) -> bool:
        """
        Returns if clamp is set on.

        Returns
        -------
        bool
            Clamp status
        """
        ...

    def get_unknown(self) -> Any:
        """
        Returns the default value when :code:`__call__` takes a
        :code:`math.nan`.

        Returns
        -------
        Any
            Unknown value
        """
        ...

    def set_interpolate(self, interpolate: Callable[[V], V]) -> TContinuousScaler:
        """
        Sets the interpolator function and returns the scaler updated.

        Parameters
        ----------
        interpolate : Callable[[V], V]
            Interpolate function

        Returns
        -------
        ContinuousScaler
            Itself
        """
        ...

    def set_clamp(self, clamp: bool) -> TContinuousScaler:
        """
        Sets the clamp status and returns the scaler updated.

        Parameters
        ----------
        clamp : bool
            Clamp status

        Returns
        -------
        ContinuousScaler
            Itself
        """
        ...

    def set_unknown(self, unknown: Any) -> TContinuousScaler:
        """
        Sets the unknown value when a :code:`math.nan` is encountered
        and returns the scaler updated.

        Parameters
        ----------
        unknown : Any
            Unknown value to set

        Returns
        -------
        ContinuousScaler
            Itself
        """
        ...


class SequentialScaler(Scaler[U, V], Generic[U, V]):
    """
    Protocol class which represents Sequential Scaler class
    """

    def set_interpolator(self, interpolator: Callable[[U], V]) -> TSequentialScaler:
        """
        Sets the interpolator function and returns the scaler updated.

        Parameters
        ----------
        interpolator : Callable[[U], V]
            Interpolator function

        Returns
        -------
        SequentialScaler
            Itself
        """
        ...


class Accessor(Generic[U, V], Protocol):
    """
    Protocol which describes how an accessor is defined.

    Accessors are found in differents modules of :code:`detroit` when a
    function iterates over an array of values and only specific values must be
    processed.

    Therefore, an accessor behaves like a function. However, the signature of
    the function can be flexible. The function can take one, two or three
    arguments.

    The first argument represents a value of the iterated array.
    The second argument represents the index of the node.
    The last and third argument is the group of nodes where :code:`node ==
    group[i]`.

    Examples
    --------

    The following functions are valid "accessors":

    >>> lambda d: d["weights"]
    >>> lambda d, i: i * d["length"]
    >>> lambda d, _, group: len(group) + d["count"]
    """

    @overload
    def __call__(self, d: U) -> V: ...

    @overload
    def __call__(self, d: U, i: int) -> V: ...

    @overload
    def __call__(self, d: U, i: int, group: list[etree.Element]) -> V: ...

    def __call__(self, *args) -> V:
        """
        Given at least one argument, returns a processed value.

        Parameters
        ----------
        d : U
            Represents a value of :code:`data` such as :code:`d = data[i]`.
        i : int
            Index of the node (i.e. :code:`node == group[i]`)
        group : list[etree.Element]
            Group of selected nodes

        Returns
        -------
        V
            A desired value
        """
        ...


class MultiPolygonGeoJSON(TypedDict):
    """
    Describes a contour as GeoJSON MultiPolygon geometry objects

    Attributes
    ----------
    type : str
        Type of contour (ex: :code:`"MultiPolygon"`)
    value : str
        Threshold value; the input values are greater than or equal to this
        value.
    coordinates : list[list[tuple[int, int]]]
        Coordinates of contours
    """

    type: str
    value: float
    coordinates: list[list[Point2D]]


class SimulationNode(TypedDict):
    """
    Describes a simulation's node.

    Attributes
    ----------
    index : int
        The node's zero-based index into nodes
    x : float
        The node's current x-position
    y : float
        The node's current y-position
    vx : float
        The node's current x-velocity
    vy : float
        The node's current y-velocity
    fx : float
        The node's fixed x-position
    fy : float
        The node's fixed y-position
    """

    index: int
    x: float
    y: float
    vx: float
    vy: float
    fx: float
    fy: float


class SimulationLink(TypedDict):
    """
    Describes a simulation's link.

    Attributes
    ----------
    source : SimulationNode
        The link's source node
    target : SimulationNode
        The link's target node
    index : int
        The zero-based index into links
    """

    source: dict
    target: dict
    index: int


SimulationNodeFunction: TypeAlias = Callable[
    [SimulationNode, int, list[SimulationNode]], T
]
SimulationLinkFunction: TypeAlias = Callable[
    [SimulationLink, int, list[SimulationLink]], T
]


class Force(Protocol):
    """
    Protocol class which represents Force object
    """

    def initialize(
        self,
        nodes: list[SimulationNode],
        random: Callable[[None], float],
    ):
        """
        Initializes the force.

        Parameters
        ----------
        nodes : list[SimulationNode]
            List of simulation nodes
        random : Callable[[None], float]
            Function which returns a random number
        """
        ...

    def __call__(self, alpha: float | None):
        """
        Applies the force on simulation nodes.

        Parameters
        ----------
        alpha : float | None
            Parameter used during the call.
        """
        ...



from .default_locale import format_default_locale, format_prefix, locale_format
from .format_specifier import format_specifier
from .precision_fixed import precision_fixed
from .precision_prefix import precision_prefix
from .precision_round import precision_round

__all__ = [
    "format_default_locale",
    "format_prefix",
    "format_specifier",
    "locale_format",
    "precision_fixed",
    "precision_prefix",
    "precision_round",
]



from .utils import EPSILON, RESULTERRBOUND, SPLITTER, estimate, sum_zerolim

CCWERRBOUNDA = (3 + 16 * EPSILON) * EPSILON
CCWERRBOUNDB = (2 + 12 * EPSILON) * EPSILON
CCWERRBOUNDC = (9 + 64 * EPSILON) * EPSILON * EPSILON

B = [None] * 4
C1 = [None] * 8
C2 = [None] * 12
D = [None] * 16
u = [None] * 4


def orient2dadapt(
    ax: float,
    ay: float,
    bx: float,
    by: float,
    cx: float,
    cy: float,
    detsum: float,
) -> float:
    acx = ax - cx
    bcx = bx - cx
    acy = ay - cy
    bcy = bx - cy

    s1 = acx * bcy
    c = SPLITTER * acx
    ahi = c - (c - acx)
    alo = acx - ahi
    c = SPLITTER * bcy
    bhi = c - (c - bcy)
    blo = bcy - bhi
    s0 = alo * blo - (s1 - ahi * bhi - alo * bhi - ahi * blo)
    t1 = acy * bcx
    c = SPLITTER * acy
    ahi = c - (c - acy)
    alo = acy - ahi
    c = SPLITTER * bcx
    bhi = c - (c - bcx)
    blo = bcx - bhi
    t0 = alo * blo - (t1 - ahi * bhi - alo * bhi - ahi * blo)
    _i = s0 - t0
    bvirt = s0 - _i
    B[0] = s0 - (_i + bvirt) + (bvirt - t0)
    _j = s1 + _i
    bvirt = _j - s1
    _0 = s1 - (_j - bvirt) + (_i - bvirt)
    _i = _0 - t1
    bvirt = _0 - _i
    B[1] = _0 - (_i + bvirt) + (bvirt - t1)
    u3 = _j + _i
    bvirt = u3 - _j
    B[2] = _j - (u3 - bvirt) + (_i - bvirt)
    B[3] = u3

    det = estimate(4, B)
    errbound = CCWERRBOUNDB * detsum
    if det >= errbound or -det >= errbound:
        return det

    bvirt = ax - acx
    acxtail = ax - (acx + bvirt) + (bvirt - cx)
    bvirt = bx - bcx
    bcxtail = bx - (bcx + bvirt) + (bvirt - cx)
    bvirt = ay - acy
    acytail = ay - (acy + bvirt) + (bvirt - cy)
    bvirt = by - bcy
    bcytail = by - (bcy + bvirt) + (bvirt - cy)

    if acxtail == 0 and acytail == 0 and bcxtail == 0 and bcytail == 0:
        return det

    errbound = CCWERRBOUNDC * detsum + RESULTERRBOUND * abs(det)
    det += (acx * bcytail + bcy * acxtail) - (acy * bcxtail + bcx * acytail)
    if det >= errbound or -det >= errbound:
        return det

    s1 = acxtail * bcy
    c = SPLITTER * acxtail
    ahi = c - (c - acxtail)
    alo = acxtail - ahi
    c = SPLITTER * bcy
    bhi = c - (c - bcy)
    blo = bcy - bhi
    s0 = alo * blo - (s1 - ahi * bhi - alo * bhi - ahi * blo)
    t1 = acytail * bcx
    c = SPLITTER * acytail
    ahi = c - (c - acytail)
    alo = acytail - ahi
    c = SPLITTER * bcx
    bhi = c - (c - bcx)
    blo = bcx - bhi
    t0 = alo * blo - (t1 - ahi * bhi - alo * bhi - ahi * blo)
    _i = s0 - t0
    bvirt = s0 - _i
    u[0] = s0 - (_i + bvirt) + (bvirt - t0)
    _j = s1 + _i
    bvirt = _j - s1
    _0 = s1 - (_j - bvirt) + (_i - bvirt)
    _i = _0 - t1
    bvirt = _0 - _i
    u[1] = _0 - (_i + bvirt) + (bvirt - t1)
    u3 = _j + _i
    bvirt = u3 - _j
    u[2] = _j - (u3 - bvirt) + (_i - bvirt)
    u[3] = u3
    C1len = sum_zerolim(4, B, 4, u, C1)

    s1 = acx * bcytail
    c = SPLITTER * acx
    ahi = c - (c - acx)
    alo = acx - ahi
    c = SPLITTER * bcytail
    bhi = c - (c - bcytail)
    blo = bcytail - bhi
    s0 = alo * blo - (s1 - ahi * bhi - alo * bhi - ahi * blo)
    t1 = acy * bcxtail
    c = SPLITTER * acy
    ahi = c - (c - acy)
    alo = acy - ahi
    c = SPLITTER * bcxtail
    bhi = c - (c - bcxtail)
    blo = bcxtail - bhi
    t0 = alo * blo - (t1 - ahi * bhi - alo * bhi - ahi * blo)
    _i = s0 - t0
    bvirt = s0 - _i
    u[0] = s0 - (_i + bvirt) + (bvirt - t0)
    _j = s1 + _i
    bvirt = _j - s1
    _0 = s1 - (_j - bvirt) + (_i - bvirt)
    _i = _0 - t1
    bvirt = _0 - _i
    u[1] = _0 - (_i + bvirt) + (bvirt - t1)
    u3 = _j + _i
    bvirt = u3 - _j
    u[2] = _j - (u3 - bvirt) + (_i - bvirt)
    u[3] = u3
    C2len = sum_zerolim(C1len, C1, 4, u, C2)

    s1 = acxtail * bcytail
    c = SPLITTER * acxtail
    ahi = c - (c - acxtail)
    alo = acxtail - ahi
    c = SPLITTER * bcytail
    bhi = c - (c - bcytail)
    blo = bcytail - bhi
    s0 = alo * blo - (s1 - ahi * bhi - alo * bhi - ahi * blo)
    t1 = acytail * bcxtail
    c = SPLITTER * acytail
    ahi = c - (c - acytail)
    alo = acytail - ahi
    c = SPLITTER * bcxtail
    bhi = c - (c - bcxtail)
    blo = bcxtail - bhi
    t0 = alo * blo - (t1 - ahi * bhi - alo * bhi - ahi * blo)
    _i = s0 - t0
    bvirt = s0 - _i
    u[0] = s0 - (_i + bvirt) + (bvirt - t0)
    _j = s1 + _i
    bvirt = _j - s1
    _0 = s1 - (_j - bvirt) + (_i - bvirt)
    _i = _0 - t1
    bvirt = _0 - _i
    u[1] = _0 - (_i + bvirt) + (bvirt - t1)
    u3 = _j + _i
    bvirt = u3 - _j
    u[2] = _j - (u3 - bvirt) + (_i - bvirt)
    u[3] = u3
    Dlen = sum_zerolim(C2len, C2, 4, u, D)

    return D[Dlen - 1]


def orient2d(
    ax: float,
    ay: float,
    bx: float,
    by: float,
    cx: float,
    cy: float,
) -> float:
    """
    Returns a positive value if the points :code:`a`, :code:`b`, and :code:`c`
    occur in counterclockwise order (:code:`c` lies to the left of the directed
    line defined by points :code:`a` and :code:`b`).

    Returns a negative value if they occur in clockwise order (:code:`c` lies
    to the right of the directed line :code:`ab`).

    Returns zero if they are collinear.

    Parameters
    ----------
    ax : float
        X-coordinate of point A
    ay : float
        Y-coordinate of point A
    bx : float
        X-coordinate of point B
    by : float
        Y-coordinate of point B
    cx : float
        X-coordinate of point C
    cy : float
        Y-coordinate of point C

    Returns
    -------
    float
        Value which indicates if the points :code:`a`, :code:`b`, and :code:`c`
        occur in counterclockwise order

    Notes
    -----

    The result is also an approximation of twice the signed area of the
    triangle defined by the three points.
    """
    detleft = (ay - cy) * (bx - cx)
    detright = (ax - cx) * (by - cy)
    det = detleft - detright

    detsum = abs(detleft + detright)
    if abs(det) >= CCWERRBOUNDA * detsum:
        return det

    return -orient2dadapt(ax, ay, bx, by, cx, cy, detsum)


def orient2dfast(
    ax: float,
    ay: float,
    bx: float,
    by: float,
    cx: float,
    cy: float,
) -> float:
    """
    Simple, approximate, non-robust versions of :code:`orient2d` predicates.
    Use when robustness isn't needed.

    Parameters
    ----------
    ax : float
        X-coordinate of point A
    ay : float
        Y-coordinate of point A
    bx : float
        X-coordinate of point B
    by : float
        Y-coordinate of point B
    cx : float
        X-coordinate of point C
    cy : float
        Y-coordinate of point C

    Returns
    -------
    float
        Value which indicates if the points :code:`a`, :code:`b`, and :code:`c`
        occur in counterclockwise order
    """
    return (ay - cy) * (bx - cx) - (ax - cx) * (by - cy)



from __future__ import annotations

from typing import Generic, overload

from ..types import Number, T
from .init import init_range
from .ordinal import ScaleOrdinal


class ScaleBand(ScaleOrdinal[T, Number], Generic[T]):
    """
    Band scales are like band scales except the output range is continuous
    and numeric. The scale divides the continuous range into uniform bands.
    Band scales are typically used for bar charts with an band or categorical
    dimension.
    """

    def __init__(self):
        super().__init__()
        self._r0 = 0
        self._r1 = 1
        self._step = None
        self._bandwidth = None
        self._round = False
        self._padding_inner = 0
        self._padding_outer = 0
        self._align = 0.5
        self._rescale()

    def __call__(self, d: T) -> Number:
        """
        Given a value from the domain, returns the corresponding value from the range.

        Parameters
        ----------
        d : T
            Input value

        Returns
        -------
        Number
            Corresponding value from the range
        """
        i = self._index.get(d)
        if i is None:
            if self._unknown is None:
                return self._unknown
            self._domain.append(d)
            i = len(self._domain) - 1
            self._index[d] = i
        length = len(self._range_vals)
        if not length:
            return None
        index = i % length
        if index >= length or index < 0:
            return None
        return self._range_vals[index]

    def _rescale(self):
        n = len(self._domain)
        reverse = self._r1 < self._r0
        start = self._r1 if reverse else self._r0
        stop = self._r0 if reverse else self._r1
        self._step = (stop - start) / max(
            1, n - self._padding_inner + self._padding_outer * 2
        )
        if self._round:
            self._step = int(self._step)
        start += (stop - start - self._step * (n - self._padding_inner)) * self._align
        self._bandwidth = self._step * (1 - self._padding_inner)
        if self._round:
            start = round(start)
            self._bandwidth = round(self._bandwidth)
        values = [start + self._step * i for i in range(n)]
        return super().set_range(values[::-1] if reverse else values)

    def set_domain(self, domain: list[T]) -> ScaleBand:
        """
        Sets the scale's domain to the specified array of values

        Parameters
        ----------
        domain : list[T]
            Domain

        Returns
        -------
        ScaleBand
            Itself
        """
        super().set_domain(domain)
        return self._rescale()

    def get_domain(self):
        return self._domain.copy()

    def set_range(self, range_vals: list[Number]) -> ScaleBand:
        """
        Sets the scale's range to the specified array of numbers

        Parameters
        ----------
        range_vals : list[Number]
            Range

        Returns
        -------
        ScaleBand
            Itself
        """
        self._r0, self._r1 = map(float, range_vals)
        return self._rescale()

    def get_range(self) -> list[Number]:
        return [self._r0, self._r1]

    def set_range_round(self, range_vals: list[int | range]) -> ScaleBand:
        """
        Sets the scale's range to the specified array of values
        and sets scale's interpolator to :code:`interpolate_round`.

        Parameters
        ----------
        range_vals : list[int | range]
            Range values

        Returns
        -------
        ScaleBand
            Itself
        """
        self._r0, self._r1 = map(float, range_vals)
        self._round = True
        return self._rescale()

    def get_bandwidth(self):
        return self._bandwidth

    def get_step(self):
        return self._step

    def set_round(self, round_val: bool) -> ScaleBand:
        """
        Enable or disable rounding accordingly

        Parameters
        ----------
        round_val : bool
            Round value

        Returns
        -------
        ScaleBand
            Itself
        """
        self._round = bool(round_val)
        return self._rescale()

    def get_round(self):
        return self._round

    def set_padding(self, padding: Number) -> ScaleBand:
        """
        A convenience method for setting the inner and outer padding
        to the same padding value.

        Parameters
        ----------
        padding : Number
            Padding value

        Returns
        -------
        ScaleBand
            Itself
        """
        self._padding_outer = float(padding)
        self._padding_inner = min(1, self._padding_outer)
        return self._rescale()

    def get_padding(self):
        return self._padding_inner

    def set_padding_inner(self, padding_inner: Number) -> ScaleBand:
        """
        Sets the inner padding to the specified number which must
        be less than or equal to 1

        Parameters
        ----------
        padding_inner : Number
            Inner padding value

        Returns
        -------
        ScaleBand
            Itself
        """
        self._padding_inner = min(1, float(padding_inner))
        return self._rescale()

    def set_padding_outer(self, padding_outer: Number) -> ScaleBand:
        """
        Sets the outer padding to the specified number
        which is typically in the range [0, 1]

        Parameters
        ----------
        padding_outer : Number
            Outer padding value

        Returns
        -------
        ScaleBand
            Itself
        """
        self._padding_outer = float(padding_outer)
        return self._rescale()

    def get_padding_inner(self):
        return self._padding_inner

    def get_padding_outer(self):
        return self._padding_outer

    def set_align(self, align: Number) -> ScaleBand:
        """
        Sets the alignment to the specified value which must
        be in the range [0, 1]

        Parameters
        ----------
        align : Number
            Alignment value

        Returns
        -------
        ScaleBand
            Itself
        """
        self._align = max(0, min(1, float(align)))
        return self._rescale()

    def get_align(self):
        return self._align

    def copy(self):
        return (
            ScaleBand()
            .set_domain(self._domain)
            .set_range([self._r0, self._r1])
            .set_round(self._round)
            .set_padding_inner(self._padding_inner)
            .set_padding_outer(self._padding_outer)
            .set_align(self._align)
        )

    def __str__(self) -> str:
        name = self.__class__.__name__
        attrbs = ["domain", "range", "padding_inner", "padding_outer"]
        attrbs = (f"{a}={getattr(self, f'get_{a}')()}" for a in attrbs)
        attrbs = ", ".join(attrbs)
        return f"{name}({attrbs})"

    def __repr__(self) -> str:
        name = self.__class__.__name__
        addr = id(self)
        return f"<{name} at {hex(addr)}>"


@overload
def scale_band() -> ScaleBand: ...


@overload
def scale_band(range_vals: list[Number]) -> ScaleBand: ...


@overload
def scale_band(domain: list[T], range_vals: list[Number]) -> ScaleBand: ...


def scale_band(*args):
    """
    Builds a new band scale with the specified domain
    and range, no padding, no rounding and center alignment

    Parameters
    ----------
    domain : list[T]
        Array of values
    range_vals : list[Number]
        Array of numbers

    Returns
    -------
    ScaleBand
        Scale object

    Examples
    --------

    >>> scale = d3.scale_band(["a", "b", "c"], [0, 960])
    >>> for c in "abcdefgh":
    ...     print(c, scale(c))
    ...
    ...
    a 0.0
    b 320.0
    c 640.0
    d 0.0
    e 320.0
    f 640.0
    g 0.0
    h 320.0
    >>> scale.get_bandwidth()
    320.0
    """
    scale = ScaleBand()
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)


@overload
def scale_point() -> ScaleBand[T]: ...


@overload
def scale_point(range_vals: list[Number]) -> ScaleBand[T]: ...


@overload
def scale_point(domain: list[T], range_vals: list[Number]) -> ScaleBand[T]: ...


def scale_point(*args):
    """
    Builds a new point scale with the specified domain and
    range, no padding, no rounding and center alignment

    Parameters
    ----------
    domain : list[T]
        Array of values
    range_vals : list[Number]
        Array of numbers

    Returns
    -------
    ScaleBand
        Scale object

    Examples
    --------

    >>> scale = d3.scale_point(["a", "b", "c"], [0, 960])
    >>> for c in "abcdefgh":
    ...     print(c, scale(c))
    ...
    ...
    a 0.0
    b 480.0
    c 960.0
    d 0.0
    e 480.0
    f 960.0
    g 0.0
    h 480.0
    >>> scale.get_bandwidth()
    0.0
    """
    scale = ScaleBand().set_padding_inner(1)
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)



from math import hypot

from ..types import Point2D


def length(polygon: list[Point2D]) -> float:
    """
    Returns the length of the perimeter of the specified polygon.

    Parameters
    ----------
    polygon : list[Point2D]
        Polygon

    Returns
    -------
    float
        Length of the perimeter of the specified polygon.

    Examples
    --------
    >>> d3.polygon_length([[1, 1], [1.5, 0], [2, 1]])
    3.23606797749979
    """
    b = polygon[-1]
    xb, yb = b
    perimeter = 0

    for point in polygon:
        xa, ya = xb, yb
        b = point
        xb, yb = point
        xa -= xb
        ya -= yb
        perimeter += hypot(xa, ya)

    return perimeter



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Protocol, TypeAlias

from ..types import GeoJSON, Point2D, Vec2D


class SpatialTransform(Protocol):
    """
    Describes a spatial transformation
    """

    def __call__(self, x: float, y: float) -> Point2D:
        """
        Transforms the x and y values into two new values.

        Parameters
        ----------
        x : float
            x value
        y : float
            y value

        Returns
        -------
        Point2D
            New values
        """
        ...

    def invert(self, x: float, y: float) -> Point2D:
        """
        Optional method which makes the invert transformation.

        Parameters
        ----------
        x : float
            x value
        y : float
            y value

        Returns
        -------
        Point2D
            Inverted values
        """
        ...


class LineStream(ABC):
    @abstractmethod
    def line_start(self):
        """
        Indicates the start of a line or ring. Within a polygon, indicates the
        start of a ring. The first ring of a polygon is the exterior ring, and
        is typically clockwise. Any subsequent rings indicate holes in the
        polygon, and are typically counterclockwise.
        """
        ...

    @abstractmethod
    def line_end(self):
        """
        Indicates the end of a line or ring. Within a polygon, indicates the
        end of a ring. Unlike GeoJSON, the redundant closing coordinate of a
        ring is not indicated via point, and instead is implied via lineEnd
        within a polygon.
        """
        ...

    @abstractmethod
    def point(self, x: float, y: float):
        """
        Indicates a point with the specified coordinates x and y (and
        optionally z). The coordinate system is unspecified and
        implementation-dependent; for example, projection streams require
        spherical coordinates in degrees as input. Outside the context of a
        polygon or line, a point indicates a point geometry object (Point or
        MultiPoint). Within a line or polygon ring, the point indicates a
        control point.

        Parameters
        ----------
        x : float
            x value
        y : float
            y value
        """
        ...

    def sphere(self):
        """
        Indicates the sphere (the globe; the unit sphere centered at (0,0,0)).
        """
        ...


class PolygonStream(LineStream):
    @abstractmethod
    def polygon_start(self):
        """
        Indicates the start of a polygon. The first line of a polygon indicates
        the exterior ring, and any subsequent lines indicate interior holes.
        """
        ...

    @abstractmethod
    def polygon_end(self):
        """
        Indicates the end of a polygon.
        """
        ...


Stream: TypeAlias = LineStream | PolygonStream


class Context(ABC):
    """
    Context definition
    """

    @abstractmethod
    def arc(self, x: float, y: float, r: float):
        """
        Adds an arc

        Parameters
        ----------
        x : float
            x value of the end point
        y : float
            y value of the end point
        r : float
            radius value
        """
        ...

    @abstractmethod
    def move_to(self, x: float, y: float):
        """
        Moves to a specific point

        Parameters
        ----------
        x : float
            x value of the end point
        y : float
            y value of the end point
        """
        ...

    @abstractmethod
    def line_to(self, x: float, y: float):
        """
        Makes a line

        Parameters
        ----------
        x : float
            x value of the end point
        y : float
            y value of the end point
        """
        ...

    @abstractmethod
    def close_path(self):
        """
        Closes the path
        """
        ...


class RawProjection(Protocol):
    """
    Raw projections are point transformation functions that are used to
    implement custom projections; they typically passed to geo_projection or
    geo_projection_mutator. They are exposed here to facilitate the derivation
    of related projections. Raw projections take spherical coordinates
    :code:`[lambda, phi]` in radians (not degrees!) and return a point
    :code:`[x, y]`, typically in the unit square centered around the origin.
    """

    def __call__(self, lambda_: float, phi: float) -> Point2D:
        """
        Projects the specified point :code:`[lambda, phi]` in radians,
        returning a new point :code:`[x, y]` in unitless coordinates.

        Parameters
        ----------
        lambda_ : float
            :math:`\\lambda` value
        phi : float
            :math:`\\phi` value

        Returns
        -------
        Point2D
            New point
        """
        ...

    def invert(self, x: float, y: float) -> Point2D:
        """
        Optional method: the inverse of the projection.

        Parameters
        ----------
        x : float
            x value
        y : float
            y value

        Returns
        -------
        Point2D
            :code:`[lambda, phi]` coordinates in radians
        """
        ...


class Projection(ABC):
    @abstractmethod
    def __call__(self, point: Point2D) -> Point2D:
        """
        Returns a new array :code:`[x, y]` (typically in pixels) representing
        the projected point of the given point. The point must be specified as
        a two-element array :code:`[longitude, latitude]` in degrees. May
        return :code:`None` if the specified point has no defined projected position,
        such as when the point is outside the clipping bounds of the
        projection.

        Parameters
        ----------
        point : Point2D
            Point :code:`[longitude, latitude]`

        Returns
        -------
        Point2D
            New projected point :code:`[x, y]`
        """
        ...

    @abstractmethod
    def invert(self, point: Point2D) -> Point2D:
        """
        Returns a new array :code:`[longitude, latitude]` in degrees
        representing the unprojected point of the given projected point. The
        point must be specified as a two-element array :code:`[x, y]`
        (typically in pixels). May return null if the specified point has no
        defined projected position, such as when the point is outside the
        clipping bounds of the projection.

        This method is only defined on invertible projections.

        Parameters
        ----------
        point : Point2D
            2D point where coordinates are in pixels

        Returns
        -------
        Point2D
            2D point where coordinates are in degrees
        """
        ...

    @abstractmethod
    def stream(self, stream: PolygonStream) -> PolygonStream:
        """
        Returns a projection stream for the specified output stream. Any input
        geometry is projected before being streamed to the output stream. A
        typical projection involves several geometry transformations: the input
        geometry is first converted to radians, rotated on three axes, clipped
        to the small circle or cut along the antimeridian, and lastly projected
        to the plane with adaptive resampling, scale and translation.

        Parameters
        ----------
        stream : PolygonStream
            Stream object

        Returns
        -------
        PolygonStream
            Modified stream object
        """
        ...

    def set_preclip(
        self, preclip: Callable[[PolygonStream], PolygonStream]
    ) -> Projection:
        """
        If preclip is specified, sets the projection's spherical clipping to
        the specified function and returns the projection; preclip is a
        function that takes a projection stream and returns a clipped stream.

        Parameters
        ----------
        preclip : Callable[[PolygonStream], PolygonStream]
            Preclip function

        Returns
        -------
        Projection
            Itself
        """
        ...

    def set_postclip(
        self, postclip: Callable[[PolygonStream], PolygonStream]
    ) -> Projection:
        """
        If postclip is specified, sets the projection's Cartesian clipping to
        the specified function and returns the projection; postclip is a
        function that takes a projection stream and returns a clipped stream.

        Parameters
        ----------
        postclip : Callable[[PolygonStream], PolygonStream]
            Postclip function

        Returns
        -------
        Projection
            Itself
        """
        ...

    def set_clip_angle(self, clip_angle: float | None = None) -> Projection:
        """
        If angle is specified, sets the projection's clipping circle radius to
        the specified angle in degrees and returns the projection. If angle is
        :code:`None`, switches to antimeridian cutting rather than small-circle
        clipping. If angle is not specified, returns the current clip angle
        which defaults to :code:`None`. Small-circle clipping is independent of
        viewport clipping via projection.clipExtent.

        Parameters
        ----------
        clip_angle : float | None
            Clip angle

        Returns
        -------
        Projection
            Itself
        """
        ...

    def set_clip_extent(
        self, clip_extent: tuple[Point2D, Point2D] | None = None
    ) -> Projection:
        """
        If extent is specified, sets the projection's viewport clip extent to
        the specified bounds in pixels and returns the projection. The extent
        bounds are specified as an array :math:`[[x_0, y_0], [x_1, y_1]]`,
        where :math:`x_0` is the left-side of the viewport, :math:`y_0` is the
        top, :math:`x_1` is the right and :math:`y_1` is the bottom. If extent
        is :code:`None`, no viewport clipping is performed. If extent is not
        specified, returns the current viewport clip extent which defaults to
        :code:`None`. Viewport clipping is independent of small-circle clipping
        via projection.clipAngle.

        Parameters
        ----------
        clip_extent : tuple[Point2D, Point2D] | None


        Returns
        -------
        Projection
            Itself
        """
        ...

    @abstractmethod
    def scale(self, k: float) -> Projection:
        """
        If scale is specified, sets the projection's scale factor to the
        specified value and returns the projection. The scale factor
        corresponds linearly to the distance between projected points; however,
        absolute scale factors are not equivalent across projections.

        Parameters
        ----------
        k : float
            Scale factor

        Returns
        -------
        Projection
            Itself
        """
        ...

    @abstractmethod
    def translate(self, translation: Vec2D) -> Projection:
        """
        If translate is specified, sets the projection's translation offset to
        the specified two-element array :code:`[tx, ty]` and returns the
        projection. The translation offset determines the pixel coordinates of
        the projection's center. The default translation offset places
        :math:`[0°,0°]` at the center of a :math:`960 \\times 500` area.

        Parameters
        ----------
        translation : Vec2D
            Translation 2D vector

        Returns
        -------
        Projection
            Itself
        """
        ...

    def set_center(self, center: Point2D) -> Projection:
        """
        If center is specified, sets the projection's center to the specified
        center, a two-element array of :code:`[longitude, latitude]` in degrees
        and returns the projection.

        Parameters
        ----------
        center : Point2D
            Center point

        Returns
        -------
        Projection
            Itself
        """
        ...

    def rotate(
        self, angles: tuple[float, float] | tuple[float, float, float]
    ) -> Projection:
        """
        If angles is specified, sets the projection's three-axis spherical
        rotation to the specified value, which must be a two- or three-element
        array of numbers :math:`[\\lambda, \\phi, \\gamma]` specifying the
        rotation angles in degrees about each spherical axis. (These correspond
        to yaw, pitch and roll).

        Parameters
        ----------
        angles : tuple[float, float] | tuple[float, float, float]
            :math:`[\\lambda, \\phi, \\gamma]` values or :math:`[\\lambda, \\phi]` values

        Returns
        -------
        Projection
            Itself
        """
        ...

    def set_angle(self, angle: float) -> Projection:
        """
        If angle is specified, sets the projection's post-projection planar
        rotation angle to the specified angle in degrees and returns the
        projection.

        Parameters
        ----------
        angle : float
            Angle value

        Returns
        -------
        Projection
            Itself
        """
        ...

    def set_reflect_x(self, reflect_x: bool) -> Projection:
        """
        If reflect is specified, sets whether or not the x-dimension is
        reflected (negated) in the output. This can be useful to display sky
        and astronomical data with the orb seen from below: right ascension
        (eastern direction) will point to the left when North is pointing up.

        Parameters
        ----------
        reflect_x : bool
            Boolean value

        Returns
        -------
        Projection
            Itself
        """
        ...

    def set_reflect_y(self, reflect_y: bool) -> Projection:
        """
        If reflect is specified, sets whether or not the y-dimension is
        reflected (negated) in the output. This is especially useful for
        transforming from standard spatial reference systems, which treat
        positive y as pointing up, to display coordinate systems such as Canvas
        and SVG, which treat positive y as pointing down.

        Parameters
        ----------
        reflect_y : bool
            Boolean value

        Returns
        -------
        Projection
            Itself
        """
        ...

    @abstractmethod
    def set_precision(self, precision: float) -> Projection:
        """
        If precision is specified, sets the threshold for the projection's
        adaptive resampling to the specified value in pixels and returns the
        projection. This value corresponds to the Douglas–Peucker distance.

        Parameters
        ----------
        precision : float
            Precision value

        Returns
        -------
        Projection
            Itself
        """
        ...

    @abstractmethod
    def fit_extent(self, extent: tuple[Point2D, Point2D], obj: GeoJSON) -> Projection:
        """
        Sets the projection's scale and translate to fit the specified GeoJSON
        object in the center of the given extent. The extent is specified as an
        array :math:`[[x_0, y_0], [x_1, y_1]]`, where :math:`x_0` is the left
        side of the bounding box, :math:`y_0` is the top, :math:`x_1`₁ is the
        right and :math:`y_1` is the bottom. Returns the projection.

        Any clip extent is ignored when determining the new scale and
        translate. The precision used to compute the bounding box of the given
        object is computed at an effective scale of 150.

        Parameters
        ----------
        extent : tuple[Point2D, Point2D]
            Extent values
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        Projection
            Itself
        """
        ...

    @abstractmethod
    def fit_size(self, size: Point2D, obj: GeoJSON) -> Projection:
        """
        A convenience method for projection.fit_extent where the top-left
        corner of the extent is [0, 0].

        Parameters
        ----------
        size : Point2D
            Size values
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        Projection
            Itself
        """
        ...

    @abstractmethod
    def fit_width(self, width: float, obj: GeoJSON) -> Projection:
        """
        A convenience method for projection.fit_size where the height is
        automatically chosen from the aspect ratio of object and the given
        constraint on width.

        Parameters
        ----------
        width : float
            Width value
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        Projection
            Itself
        """
        ...

    @abstractmethod
    def fit_height(self, height: float, obj: GeoJSON) -> Projection:
        """
        A convenience method for projection.fit_size where the width is
        automatically chosen from the aspect ratio of object and the given
        constraint on height.

        Parameters
        ----------
        height : float
            Height value
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        Projection
            Itself
        """
        ...



from ..colors import colors
from ..ramp import ramp

scheme = [
    "e0f3dba8ddb543a2ca",
    "f0f9e8bae4bc7bccc42b8cbe",
    "f0f9e8bae4bc7bccc443a2ca0868ac",
    "f0f9e8ccebc5a8ddb57bccc443a2ca0868ac",
    "f0f9e8ccebc5a8ddb57bccc44eb3d32b8cbe08589e",
    "f7fcf0e0f3dbccebc5a8ddb57bccc44eb3d32b8cbe08589e",
    "f7fcf0e0f3dbccebc5a8ddb57bccc44eb3d32b8cbe0868ac084081",
]
SCHEME_GNBU = list(map(colors, scheme))
interpolate_gnbu = ramp(SCHEME_GNBU)



from typing import Any

from lxml import etree

from ..array import argpass
from ..types import Accessor, EtreeFunction, Number, T


def tostring(value: list[Any] | Any) -> str:
    """
    If :code:`value` is a list of elements, the function converts elements
    strings, concatenate them into a single string and return it.
    If :code:`value` is a single element, the function converts it into a
    string.

    Parameters
    ----------
    value : list[Any] | Any
        Value to convert into string

    Returns
    -------
    str
        Output
    """
    return " ".join(map(str, value)) if isinstance(value, list) else str(value)


def attr_constant(name: str, value: list[Any] | Any) -> EtreeFunction[T, None]:
    """
    Returns a function which adds an attribute to nodes given a constant value.

    Parameters
    ----------
    name : str
        Attribute name
    value : list[Any] | Any
        Constant value of the attribute

    Returns
    -------
    EtreeFunction[T, None]
        Function which adds an attribute to nodes
    """
    value = tostring(value)

    def callback(node: etree.Element, data: T, i: int, group: list[etree.Element]):
        node.set(name, value)

    return callback


def attr_function(
    name: str, value: Accessor[T, str | Number]
) -> EtreeFunction[T, None]:
    """
    Returns a function which adds an attribute to nodes based on an accessor
    function.

    Parameters
    ----------
    name : str
        Attribute name
    value : Accessor[T, str | Number]
        Accessor function

    Returns
    -------
    EtreeFunction[T, None]
        Function which adds an attribute to nodes
    """
    value = argpass(value)

    def callback(node: etree.Element, data: T, i: int, group: list[etree.Element]):
        node.set(name, tostring(value(data, i, group)))

    return callback



from abc import ABC, abstractmethod


class Context(ABC):
    """
    Context definition
    """

    @abstractmethod
    def arc(
        self,
        x: float,
        y: float,
        r: float,
        start_angle: float,
        end_angle: float,
    ):
        """
        Adds an arc

        Parameters
        ----------
        x : float
            x value of the end point
        y : float
            y value of the end point
        r : float
            radius value
        start_angle : float
            start angle value
        end_angle : float
            end angle value
        """
        ...

    @abstractmethod
    def rect(self, x: float, y: float, w: float, h: float):
        """
        Adds an rectangle

        Parameters
        ----------
        x : float
            x value of the top left corner of the rectangle
        y : float
            y value of the top left corner of the rectangle
        w : float
            width of the rectangle
        h : float
            height of the rectangle
        """
        ...

    @abstractmethod
    def move_to(self, x: float, y: float):
        """
        Moves to a specific point

        Parameters
        ----------
        x : float
            x value of the end point
        y : float
            y value of the end point
        """
        ...

    @abstractmethod
    def line_to(self, x: float, y: float):
        """
        Makes a line

        Parameters
        ----------
        x : float
            x value of the end point
        y : float
            y value of the end point
        """
        ...

    @abstractmethod
    def close_path(self):
        """
        Closes the path
        """
        ...



from collections.abc import Callable

from ..types import T


def constant(x: T) -> Callable[..., T]:
    """
    Returns a function which returns the same value :code:`x` constantly.

    Parameters
    ----------
    x : T
        Value to return constantly

    Returns
    -------
    Callable[..., T]
        Function which returns :code:`x` whatever are arguments
    """

    def f(*args: ...) -> T:
        return x

    return f



from ..colors import colors
from ..ramp import ramp

scheme = [
    "ece7f2a6bddb2b8cbe",
    "f1eef6bdc9e174a9cf0570b0",
    "f1eef6bdc9e174a9cf2b8cbe045a8d",
    "f1eef6d0d1e6a6bddb74a9cf2b8cbe045a8d",
    "f1eef6d0d1e6a6bddb74a9cf3690c00570b0034e7b",
    "fff7fbece7f2d0d1e6a6bddb74a9cf3690c00570b0034e7b",
    "fff7fbece7f2d0d1e6a6bddb74a9cf3690c00570b0045a8d023858",
]
SCHEME_PUBU = list(map(colors, scheme))
interpolate_pubu = ramp(SCHEME_PUBU)



import math

from ...selection import Selection
from .common import Curve, isvaluable


class BezierTrait:
    def _bezier_curve_to(self, x, y):
        self._context.bezier_curve_to(
            (2 * self._x0 + self._x1) / 3,
            (2 * self._y0 + self._y1) / 3,
            (self._x0 + 2 * self._x1) / 3,
            (self._y0 + 2 * self._y1) / 3,
            (self._x0 + 4 * self._x1 + x) / 6,
            (self._y0 + 4 * self._y1 + y) / 6,
        )


class BasisCurve(Curve, BezierTrait):
    def __init__(self, context):
        self._context = context
        self._line = math.nan
        self._x0 = math.nan
        self._y0 = math.nan
        self._x1 = math.nan
        self._y1 = math.nan

    def area_start(self):
        self._line = 0

    def area_end(self):
        self._line = math.nan

    def line_start(self):
        self._x0 = math.nan
        self._x1 = math.nan
        self._y0 = math.nan
        self._y1 = math.nan
        self._point = 0

    def line_end(self):
        if self._point == 3:
            self._bezier_curve_to(self._x1, self._y1)
            self._context.line_to(self._x1, self._y1)
        elif self._point == 2:
            self._context.line_to(self._x1, self._y1)

        if isvaluable(self._line) or (self._line != 0 and self._point == 1):
            self._context.close_path()
        self._line = 1 - self._line

    def point(self, x, y):
        if self._point == 0:
            self._point = 1
            if isvaluable(self._line):
                self._context.line_to(x, y)
            else:
                self._context.move_to(x, y)
        elif self._point == 1:
            self._point = 2
        elif self._point == 2:
            self._point = 3
            self._context.line_to(
                (5 * self._x0 + self._x1) / 6, (5 * self._y0 + self._y1) / 6
            )
            self._bezier_curve_to(x, y)
        else:
            self._bezier_curve_to(x, y)
        self._x0 = self._x1
        self._x1 = x
        self._y0 = self._y1
        self._y1 = y


def curve_basis(context: Selection) -> Curve:
    """
    Produces a cubic basis spline using the specified control points. The first
    and last points are triplicated such that the spline starts at the first
    point and ends at the last point, and is tangent to the line between the
    first and second points, and to the line between the penultimate and last
    points.

    Parameters
    ----------
    context : Selection
        Context

    Returns
    -------
    Curve
        Curve object

    Examples
    --------

    >>> points = [
    ...     [2, 2],
    ...     [6, 8],
    ...     [10, 10],
    ...     [12, 10],
    ...     [14, 4],
    ...     [20, 4],
    ...     [24, 8],
    ...     [29, 6],
    ...     [32, 4],
    ...     [35, 5],
    ...     [38, 2],
    ... ]
    >>> line = d3.line().curve(d3.curve_basis)
    >>> line(points)
    'M2,2L2.667,3C3.333,4,4.667,6,6,7.333C7.333,8.667,8.667,9.333,9.667,9.667C10.667,10,11.333,10,12,9C12.667,8,13.333,6,14.667,5C16,4,18,4,19.667,4.667C21.333,5.333,22.667,6.667,24.167,7C25.667,7.333,27.333,6.667,28.667,6C30,5.333,31,4.667,32,4.500C33,4.333,34,4.667,35,4.333C36,4,37,3,37.500,2.500L38,2'

    **Result**

    .. image:: ../../figures/light_curve_basis.svg
       :align: center
       :class: only-light

    .. image:: ../../figures/dark_curve_basis.svg
       :align: center
       :class: only-dark
    """
    return BasisCurve(context)



from math import asin, atan2, cos, nan, sin, sqrt

from ..types import Point3D, Vec3D


def spherical(cartesian: Point3D) -> Point3D:
    return [atan2(cartesian[1], cartesian[0]), asin(cartesian[2])]


def cartesian(spherical: Point3D) -> Point3D:
    lambda_ = spherical[0]
    phi = spherical[1]
    cos_phi = cos(phi)
    return [cos_phi * cos(lambda_), cos_phi * sin(lambda_), sin(phi)]


def cartesian_dot(a: Vec3D, b: Vec3D) -> float:
    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2]


def cartesian_cross(a: Vec3D, b: Vec3D) -> Vec3D:
    return [
        a[1] * b[2] - a[2] * b[1],
        a[2] * b[0] - a[0] * b[2],
        a[0] * b[1] - a[1] * b[0],
    ]


def cartesian_add_in_place(a: Vec3D, b: Vec3D):
    a[0] += b[0]
    a[1] += b[1]
    a[2] += b[2]


def cartesian_scale(vector: Vec3D, k: float) -> Vec3D:
    return [vector[0] * k, vector[1] * k, vector[2] * k]


def cartesian_normalize_in_place(d: Vec3D):
    length = sqrt(d[0] * d[0] + d[1] * d[1] + d[2] * d[2])
    if length == 0.0:
        return [nan, nan, nan]
    d[0] /= length
    d[1] /= length
    d[2] /= length
